import numpy as np

from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from astropy.wcs import WCS

from ...spectra import Spectrum1D
from ..registers import data_loader, custom_writer
from ..parsing_utils import (generic_spectrum_from_table,
                             spectrum_from_column_mapping,
                             read_fileobj_or_hdulist)

__all__ = ['tabular_fits_loader', 'tabular_fits_writer']


def identify_tabular_fits(origin, *args, **kwargs):
    # args[0] = filename
    hdu = kwargs.get('hdu', 1)
    # Check if filename conforms to naming convention and writer has not been
    # asked to write to primary (IMAGE-only) HDU
    if origin == 'write':
        return (hdu > 0 and args[0].endswith(('.fits', '.fit')) and not
                args[0].endswith(('wcs.fits', 'wcs1d.fits', 'wcs.fit')))

    # Test if fits has extension of type BinTable and check against
    # known keys of already defined specific formats
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return (len(hdulist) > 1 and
                isinstance(hdulist[hdu], fits.BinTableHDU) and not
                (hdulist[0].header.get('TELESCOP') == 'MULTI' and
                hdulist[0].header.get('HLSPACRN') == 'MUSCLES' and
                hdulist[0].header.get('PROPOSID') == 13650) and not
                (hdulist[0].header.get('TELESCOP') == 'SDSS 2.5-M' and
                hdulist[0].header.get('FIBERID') > 0) and not
                (hdulist[0].header.get('TELESCOP') == 'HST' and
                hdulist[0].header.get('INSTRUME') in ('COS', 'STIS')) and not
                hdulist[0].header.get('TELESCOP') == 'JWST')


@data_loader("tabular-fits", identifier=identify_tabular_fits,
             dtype=Spectrum1D, extensions=['fits', 'fit'], priority=6)
def tabular_fits_loader(file_obj, column_mapping=None, hdu=1, **kwargs):
    """
    Load spectrum from a FITS file.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
            FITS file name, object (provided from name by Astropy I/O Registry),
            or HDUList (as resulting from astropy.io.fits.open()).
    hdu: int
        The HDU of the fits file (default: 1st extension) to read from
    column_mapping : dict
        A dictionary describing the relation between the FITS file columns
        and the arguments of the `Spectrum1D` class, along with unit
        information. The dictionary keys should be the FITS file column names
        while the values should be a two-tuple where the first element is the
        associated `Spectrum1D` keyword argument, and the second element is the
        unit for the ASCII file column::

            column_mapping = {'FLUX': ('flux', 'Jy')}

    Returns
    -------
    data: Spectrum1D
        The spectrum that is represented by the data in this table.
    """
    # Parse the wcs information. The wcs will be passed to the column finding
    # routines to search for spectral axis information in the file.
    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        wcs = WCS(hdulist[hdu].header)

    tab = Table.read(file_obj, format='fits', hdu=hdu)

    # Minimal checks for wcs consistency with table data -
    # assume 1D spectral axis (having shape (0, NAXIS1),
    # or alternatively compare against shape of 1st column.
    if not (wcs.naxis == 1 and wcs.array_shape[-1] == len(tab) or
            wcs.array_shape == tab[tab.colnames[0]].shape):
        wcs = None

    # If no column mapping is given, attempt to parse the file using
    # unit information
    if column_mapping is None:
        return generic_spectrum_from_table(tab, wcs=wcs, **kwargs)

    return spectrum_from_column_mapping(tab, column_mapping, wcs=wcs)


@custom_writer("tabular-fits")
def tabular_fits_writer(spectrum, file_name, hdu=1, update_header=False, **kwargs):
    """
    Write spectrum to BINTABLE extension of a FITS file.

    Parameters
    ----------
    spectrum: Spectrum1D
    file_name: str
        The path to the FITS file
    hdu: int
        Header Data Unit in FITS file to write to (currently only extension HDU 1)
    update_header: bool
        Update FITS header with all compatible entries in `spectrum.meta`
    wunit : str or `~astropy.units.Unit`
        Unit for the spectral axis (wavelength or frequency-like)
    funit : str or `~astropy.units.Unit`
        Unit for the flux (and associated uncertainty)
    wtype : str or `~numpy.dtype`
        Floating point type for storing spectral axis array
    ftype : str or `~numpy.dtype`
        Floating point type for storing flux array
    """
    if hdu < 1:
        raise ValueError(f'FITS does not support BINTABLE extension in HDU {hdu}.')

    header = spectrum.meta.get('header', fits.header.Header()).copy()

    if update_header:
        hdr_types = (str, int, float, complex, bool,
                     np.floating, np.integer, np.complexfloating, np.bool_)
        header.update([keyword for keyword in spectrum.meta.items() if
                       isinstance(keyword[1], hdr_types)])

    # Strip header of FITS reserved keywords
    for keyword in ['NAXIS', 'NAXIS1', 'NAXIS2']:
        header.remove(keyword, ignore_missing=True)

    # Add dispersion array and unit
    wtype = kwargs.pop('wtype', spectrum.spectral_axis.dtype)
    wunit = u.Unit(kwargs.pop('wunit', spectrum.spectral_axis.unit))
    disp = spectrum.spectral_axis.to(wunit, equivalencies=u.spectral())

    # Mapping of spectral_axis types to header TTYPE1 (no "torque/work" types!)
    dispname = str(wunit.physical_type)
    if dispname == "length":
        dispname = "wavelength"
    elif "energy" in dispname:
        dispname = "energy"

    # Add flux array and unit
    ftype = kwargs.pop('ftype', spectrum.flux.dtype)
    funit = u.Unit(kwargs.pop('funit', spectrum.flux.unit))
    flux = spectrum.flux.to(funit, equivalencies=u.spectral_density(disp))

    columns = [disp.astype(wtype), flux.astype(ftype)]
    colnames = [dispname, "flux"]

    # Include uncertainty - units to be inferred from spectrum.flux
    if spectrum.uncertainty is not None:
        unc = spectrum.uncertainty.quantity.to(funit, equivalencies=u.spectral_density(disp))
        columns.append(unc.astype(ftype))
        colnames.append("uncertainty")

    # For > 1D data transpose from row-major format
    for c in range(1, len(columns)):
        if columns[c].ndim > 1:
            columns[c] = columns[c].T

    tab = Table(columns, names=colnames, meta=header)

    # Todo: support writing to other HDUs than the default (1st)
    #       and an 'update' mode so different HDUs can be written to separately
    tab.write(file_name, format="fits", **kwargs)



import numpy as np
import pytest

import astropy.units as u
from astropy.nddata import StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose

from ..spectra import Spectrum1D, SpectralRegion
from ..manipulation import extract_region, extract_bounding_spectral_region, spectral_slab
from ..manipulation.utils import linear_exciser

FLUX_ARRAY = [1605.71612173, 1651.41650744, 2057.65798618, 2066.73502361, 1955.75832537,
              1670.52711471, 1491.10034446, 1637.08084112, 1471.28982259, 1299.19484483,
              1423.11195734, 1226.74494917, 1572.31888312, 1311.50503403, 1474.05051673,
              1335.39944397, 1420.61880528, 1433.18623759, 1290.26966668, 1605.67341284,
              1528.52281708, 1592.74392861, 1568.74162534, 1435.29407808, 1536.68040935,
              1157.33825995, 1136.12679394, 999.92394692, 1038.61546167, 1011.60297294]


def test_region_simple(simulated_spectra):
    np.random.seed(42)

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    region = SpectralRegion(0.6*u.um, 0.8*u.um)

    sub_spectrum = extract_region(spectrum, region)

    sub_spectrum_flux_expected = np.array(FLUX_ARRAY)

    assert_quantity_allclose(sub_spectrum.flux.value, sub_spectrum_flux_expected)


def test_slab_simple(simulated_spectra):
    np.random.seed(42)

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    sub_spectrum = spectral_slab(spectrum, 0.6*u.um, 0.8*u.um)

    sub_spectrum_flux_expected = np.array(FLUX_ARRAY)

    assert_quantity_allclose(sub_spectrum.flux.value, sub_spectrum_flux_expected)


def test_slab_pixels():
    range_iter = range(5)
    spectrum = Spectrum1D(flux=np.stack(
        [np.zeros((2, 2)) + i for i in range_iter], axis=-1) * u.nJy)
    for i in range_iter:
        sub_spectrum = spectral_slab(spectrum, i * u.pix, (i + 0.5) * u.pix)
        assert_quantity_allclose(sub_spectrum.flux, i * u.nJy)
        assert_quantity_allclose(sub_spectrum.spectral_axis, i * u.pix)


def test_region_ghz(simulated_spectra):
    spectrum = Spectrum1D(flux=simulated_spectra.s1_um_mJy_e1.flux,
                          spectral_axis=simulated_spectra.s1_um_mJy_e1.frequency)

    region = SpectralRegion(499654.09666667*u.GHz, 374740.5725*u.GHz)

    sub_spectrum = extract_region(spectrum, region)

    sub_spectrum_flux_expected = FLUX_ARRAY * u.mJy

    assert_quantity_allclose(sub_spectrum.flux, sub_spectrum_flux_expected)


def test_region_simple_check_ends(simulated_spectra):
    np.random.seed(42)

    spectrum = Spectrum1D(spectral_axis=np.linspace(1, 25, 25)*u.um, flux=np.random.random(25)*u.Jy)
    region = SpectralRegion(8*u.um, 15*u.um)

    sub_spectrum = extract_region(spectrum, region)

    assert sub_spectrum.spectral_axis.value[0] == 8
    assert sub_spectrum.spectral_axis.value[-1] == 15

    region = SpectralRegion(0*u.um, 15*u.um)
    sub_spectrum = extract_region(spectrum, region)
    assert sub_spectrum.spectral_axis.value[0] == 1

    region = SpectralRegion(8*u.um, 30*u.um)
    sub_spectrum = extract_region(spectrum, region)
    assert sub_spectrum.spectral_axis.value[-1] == 25


def test_region_empty(simulated_spectra):
    np.random.seed(42)

    empty_spectrum = Spectrum1D(spectral_axis=[]*u.um, flux=[]*u.Jy)

    # Region past upper range of spectrum
    spectrum = Spectrum1D(spectral_axis=np.linspace(1, 25, 25)*u.um, flux=np.random.random(25)*u.Jy)
    region = SpectralRegion(28*u.um, 30*u.um)
    sub_spectrum = extract_region(spectrum, region)

    assert np.allclose(sub_spectrum.spectral_axis.value, empty_spectrum.spectral_axis.value)
    assert sub_spectrum.spectral_axis.unit == empty_spectrum.spectral_axis.unit

    assert np.allclose(sub_spectrum.flux.value, empty_spectrum.flux.value)
    assert sub_spectrum.flux.unit == empty_spectrum.flux.unit

    # Region below lower range of spectrum
    spectrum = Spectrum1D(spectral_axis=np.linspace(1, 25, 25)*u.um, flux=np.random.random(25)*u.Jy)
    region = SpectralRegion(0.1*u.um, 0.3*u.um)
    sub_spectrum = extract_region(spectrum, region)

    assert np.allclose(sub_spectrum.spectral_axis.value, empty_spectrum.spectral_axis.value)
    assert sub_spectrum.spectral_axis.unit == empty_spectrum.spectral_axis.unit

    assert np.allclose(sub_spectrum.flux.value, empty_spectrum.flux.value)
    assert sub_spectrum.flux.unit == empty_spectrum.flux.unit

    # Region below lower range of spectrum and upper range in the spectrum.
    spectrum = Spectrum1D(spectral_axis=np.linspace(1, 25, 25)*u.um, flux=2*np.linspace(1, 25, 25)*u.Jy)
    region = SpectralRegion(0.1*u.um, 3.3*u.um)
    sub_spectrum = extract_region(spectrum, region)

    assert np.allclose(sub_spectrum.spectral_axis.value, [1, 2, 3])
    assert sub_spectrum.spectral_axis.unit == empty_spectrum.spectral_axis.unit

    assert np.allclose(sub_spectrum.flux.value, [2, 4, 6])
    assert sub_spectrum.flux.unit == empty_spectrum.flux.unit

    # Region has lower and upper bound the same
    with pytest.raises(Exception):
        region = SpectralRegion(3*u.um, 3*u.um)


def test_region_descending(simulated_spectra):
    np.random.seed(42)

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    region = SpectralRegion(0.8*u.um, 0.6*u.um)

    sub_spectrum = extract_region(spectrum, region)

    sub_spectrum_flux_expected = np.array(FLUX_ARRAY)

    assert_quantity_allclose(sub_spectrum.flux.value, sub_spectrum_flux_expected)


def test_descending_spectral_axis(simulated_spectra):
    spectrum = simulated_spectra.s1_um_mJy_e1_desc

    sub_spectrum_flux_expected = np.array(FLUX_ARRAY[::-1])

    region = SpectralRegion(0.8*u.um, 0.6*u.um)
    sub_spectrum = extract_region(spectrum, region)

    assert_quantity_allclose(sub_spectrum.flux.value, sub_spectrum_flux_expected)

    region = SpectralRegion(0.6*u.um, 0.8*u.um)
    sub_spectrum = extract_region(spectrum, region)

    assert_quantity_allclose(sub_spectrum.flux.value, sub_spectrum_flux_expected)


def test_region_two_sub(simulated_spectra):
    np.random.seed(42)

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    region = SpectralRegion([(0.6*u.um, 0.8*u.um), (0.86*u.um, 0.89*u.um)])

    sub_spectra = extract_region(spectrum, region)

    # Confirm the end points of the subspectra are correct
    assert_quantity_allclose(sub_spectra[0].spectral_axis[[0, -1]],
                       [0.6035353535353536, 0.793939393939394]*u.um)

    assert_quantity_allclose(sub_spectra[1].spectral_axis[[0, -1]],
                       [0.8661616161616162, 0.8858585858585859]*u.um)

    sub_spectrum_0_flux_expected = FLUX_ARRAY * u.mJy

    sub_spectrum_1_flux_expected = [1337.65312465, 1263.48914109,
                                    1589.81797876, 1548.46068415]*u.mJy

    assert_quantity_allclose(sub_spectra[0].flux, sub_spectrum_0_flux_expected)
    assert_quantity_allclose(sub_spectra[1].flux, sub_spectrum_1_flux_expected)

    # also ensure this works if the multi-region is expressed as a single
    # Quantity
    region2 = SpectralRegion([(0.6, 0.8), (0.86, 0.89)]*u.um)
    sub_spectra2 = extract_region(spectrum, region2)
    assert_quantity_allclose(sub_spectra[0].flux, sub_spectra2[0].flux)
    assert_quantity_allclose(sub_spectra[1].flux, sub_spectra2[1].flux)

    # Check that the return_single_spectrum argument works properly
    concatenated_spectrum = extract_region(spectrum, region2, return_single_spectrum=True)
    assert concatenated_spectrum.flux.shape == (34,)
    assert np.all(concatenated_spectrum.flux[0:30] == sub_spectra2[0].flux)
    assert np.all(concatenated_spectrum.flux[30:34] == sub_spectra2[1].flux)


def test_bounding_region(simulated_spectra):
    np.random.seed(42)

    spectrum = simulated_spectra.s1_um_mJy_e1
    uncertainty = StdDevUncertainty(0.1*np.random.random(len(spectrum.flux))*u.mJy)
    spectrum.uncertainty = uncertainty

    region = SpectralRegion([(0.6*u.um, 0.8*u.um), (0.86*u.um, 0.89*u.um)])

    extracted_spectrum = extract_bounding_spectral_region(spectrum, region)

    # Confirm the end points are correct
    assert_quantity_allclose(extracted_spectrum.spectral_axis[[0, -1]],
                       [0.6035353535353536, 0.8858585858585859]*u.um)

    flux_expected = (FLUX_ARRAY + [948.81864554, 1197.84859443, 1069.75268943,
                                   1118.27269184, 1301.7695563, 1206.62880648,
                                   1518.16549319, 1256.84259015, 1638.76791267,
                                   1562.05642302, 1337.65312465, 1263.48914109,
                                   1589.81797876, 1548.46068415])*u.mJy

    assert_quantity_allclose(extracted_spectrum.flux, flux_expected)

    # also ensure this works if the multi-region is expressed as a single
    # Quantity
    region2 = SpectralRegion([(0.6, 0.8), (0.86, 0.89)]*u.um)
    extracted_spectrum2 = extract_bounding_spectral_region(spectrum, region2)
    assert_quantity_allclose(extracted_spectrum2.spectral_axis[[0, -1]],
                       [0.6035353535353536, 0.8858585858585859]*u.um)
    assert_quantity_allclose(extracted_spectrum2.flux, flux_expected)


def test_extract_region_pixels():
    spectrum = Spectrum1D(spectral_axis=np.linspace(4000, 10000, 25)*u.AA,
                          flux=np.arange(25)*u.Jy)
    region = SpectralRegion(10*u.pixel, 12*u.pixel)

    extracted = extract_region(spectrum, region)

    assert_quantity_allclose(extracted.flux, [10, 11]*u.Jy)


def test_extract_region_mismatched_units():
    spectrum = Spectrum1D(spectral_axis=np.arange(25)*u.nm,
                          flux=np.arange(25)*u.Jy)

    region = SpectralRegion(100*u.AA, 119*u.AA)

    extracted = extract_region(spectrum, region)

    assert_quantity_allclose(extracted.flux, [10, 11]*u.Jy)


@pytest.mark.filterwarnings('ignore:A SpectralRegion with multiple subregions was provided')
def test_linear_excise_invert_from_spectrum():
    spec = Spectrum1D(flux=np.random.sample(100) * u.Jy,
                      spectral_axis=np.arange(100) * u.AA)
    inc_regs = (SpectralRegion(0 * u.AA, 50 * u.AA) +
                SpectralRegion(60 * u.AA, 80 * u.AA) +
                SpectralRegion(90 * u.AA, 110 * u.AA))
    exc_regs = inc_regs.invert_from_spectrum(spec)

    excised_spec = linear_exciser(spec, exc_regs)

    assert_quantity_allclose(np.diff(excised_spec[50:60].flux),
                             np.diff(excised_spec[51:61].flux))
    assert_quantity_allclose(np.diff(excised_spec[80:90].flux),
                             np.diff(excised_spec[81:91].flux))


def test_extract_masked():
    wl = [1, 2, 3, 4]*u.nm
    flux = np.arange(4)*u.Jy
    mask = [False, False, True, True]

    masked_spec = Spectrum1D(spectral_axis=wl, flux=flux, mask=mask)
    region = SpectralRegion(1.5 * u.nm, 3.5 * u.nm)

    extracted = extract_region(masked_spec, region)

    assert np.all(extracted.mask == [False, True])
    assert np.all(extracted.flux.value == [1, 2])


def test_extract_multid_flux():
    flux = np.random.sample((10, 49)) * 100
    spec = Spectrum1D(spectral_axis=np.arange(1, 50) * u.nm,
                      flux=flux * u.Jy)

    region = SpectralRegion(10 * u.nm, 20 * u.nm)
    extracted = extract_region(spec, region)

    assert extracted.shape == (10, 11)
    assert extracted[0,0].flux == spec[0,9].flux


def test_slab_multid_flux():
    flux = np.random.sample((10, 49)) * 100
    spec = Spectrum1D(spectral_axis=np.arange(1, 50) * u.nm,
                      flux=flux * u.Jy)

    extracted = spectral_slab(spec, 10 * u.nm, 20 * u.nm)

    assert extracted.shape == (10, 11)
    assert extracted[0,0].flux == spec[0,9].flux



from asdf.types import CustomType, ExtensionTypeMeta


_specutils_types = set()


class SpecutilsTypeMeta(ExtensionTypeMeta):
    """
    Keeps track of `SpecutilsType` subclasses that are created so that they can
    be stored automatically by specutils extensions for ASDF.
    """
    def __new__(mcls, name, bases, attrs):
        cls = super().__new__(mcls, name, bases, attrs)
        # Classes using this metaclass are automatically added to the list of
        # specutils extensions
        _specutils_types.add(cls)
        return cls


class SpecutilsType(CustomType, metaclass=SpecutilsTypeMeta):
    """
    Parent class of all specutils tag implementations used by ASDF
    """
    organization = 'astropy.org'
    standard = 'specutils'



import astropy.units as u
from astropy.modeling.polynomial import Chebyshev1D
from astropy.modeling.fitting import LevMarLSQFitter

from ..fitting import fit_lines
from ..manipulation.smoothing import median_smooth
from ..spectra import SpectralRegion


__all__ = ['fit_continuum', 'fit_generic_continuum']


def fit_generic_continuum(spectrum, median_window=3, model=Chebyshev1D(3),
                          fitter=LevMarLSQFitter(),
                          exclude_regions=None, weights=None):
    """
    Basic fitting of the continuum of an input spectrum. The input
    spectrum is smoothed using a median filter to remove the spikes.

    Parameters
    ----------
    spectrum : Spectrum1D
        The spectrum object overwhich the equivalent width will be calculated.
    model : list of `~astropy.modeling.Model`
        The list of models that contain the initial guess.
    median_window : float
        The width of the median smoothing kernel used to filter the data before
        fitting the continuum. See the ``kernel_size`` parameter of
        `~scipy.signal.medfilt` for more information.
    fitter : `~astropy.fitting._FitterMeta`
        The astropy fitter to use for fitting the model.
        Default: `~astropy.modeling.fitting.LevMarLSQFitter`
    exclude_regions : list of 2-tuples
        List of regions to exclude in the fitting. Passed through
        to the fitmodels routine.
    weights : list  (NOT IMPLEMENTED YET)
        List of weights to define importance of fitting regions.

    Returns
    -------
    continuum_model
        Fitted continuum as a model of whatever class ``model`` provides.

    Notes
    -----
       * Could add functionality to set the bounds in
         ``model`` if they are not set.
       * The models in the list of ``model`` are added together and passed as a
         compound model to the `~astropy.modeling.fitting.Fitter` class instance.

    """
    # Simple median smooth to remove spikes and peaks
    spectrum_smoothed = median_smooth(spectrum, median_window)

    return fit_continuum(spectrum_smoothed, model=model, fitter=fitter,
                         exclude_regions=exclude_regions, weights=weights)


def fit_continuum(spectrum, model=Chebyshev1D(3), fitter=LevMarLSQFitter(),
                  exclude_regions=None, window=None, weights=None):
    """
    Entry point for fitting using the `~astropy.modeling.fitting`
    machinery.

    Parameters
    ----------
    spectrum : Spectrum1D
        The spectrum object overwhich the equivalent width will be calculated.
    model: list of `~astropy.modeling.Model`
        The list of models that contain the initial guess.
    fitter : `~astropy.fitting._FitterMeta`
        The astropy fitter to use for fitting the model.
        Default: `~astropy.modeling.fitting.LevMarLSQFitter`
    exclude_regions : list of 2-tuples
        List of regions to exclude in the fitting. Passed through
        to the fitmodels routine.
    window : tuple of wavelengths
        Start and end wavelengths used for fitting.
    weights : list  (NOT IMPLEMENTED YET)
        List of weights to define importance of fitting regions.

    Returns
    -------
    models : list of `~astropy.modeling.Model`
        The list of models that contain the fitted model parameters.
    """
    if weights is not None:
        raise NotImplementedError("Weights support is not yet implemented.")

    w = window
    if type(w) in [list, tuple] and all([_is_valid_sequence(x) for x in w]):
        w = SpectralRegion(w)

    # Fit the flux to the model
    continuum_spectrum = fit_lines(spectrum, model, fitter, exclude_regions,
                                   weights, w)

    return continuum_spectrum


# Checks for sequences of of 2-tuples with Quantities
def _is_valid_sequence(value):
    if type(value) in [list, tuple]:
        return len(value) == 2 and \
               isinstance(value[0], u.Quantity) and \
               isinstance(value[1], u.Quantity)

    return False



"""
A module for analysis tools dealing with uncertainties or error analysis in
spectra.
"""

import numpy as np
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, InverseVariance
from ..spectra import SpectralRegion
from ..manipulation import extract_region

__all__ = ['snr', 'snr_derived']


def snr(spectrum, region=None):
    """
    Calculate the mean S/N of the spectrum based on the flux and uncertainty
    in the spectrum. This will be calculated over the regions, if they
    are specified.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the equivalent width will be calculated.

    region: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the SNR.

    Returns
    -------
    snr : `~astropy.units.Quantity` or list (based on region input)
        Signal to noise ratio of the spectrum or within the regions

    Notes
    -----
    The spectrum will need to have the uncertainty defined in order for the SNR
    to be calculated. If the goal is instead signal to noise *per pixel*, this
    should be computed directly as ``spectrum.flux / spectrum.uncertainty``.

    """

    if not hasattr(spectrum, 'uncertainty') or spectrum.uncertainty is None:
        raise Exception("Spectrum1D currently requires the uncertainty be defined.")

    # No region, therefore whole spectrum.
    if region is None:
        return _snr_single_region(spectrum)

    # Single region
    elif isinstance(region, SpectralRegion):
        return _snr_single_region(spectrum, region=region)

    # List of regions
    elif isinstance(region, list):
        return [_snr_single_region(spectrum, region=reg)
                for reg in region]


def _snr_single_region(spectrum, region=None):
    """
    Calculate the mean S/N of the spectrum based on the flux and uncertainty
    in the spectrum.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the equivalent width will be calculated.

    region: `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the SNR.

    Returns
    -------
    snr : `~astropy.units.Quantity` or list (based on region input)
        Signal to noise ratio of the spectrum or within the regions

    Notes
    -----
    This is a helper function for the above `snr()` method.

    """

    if region is not None:
        calc_spectrum = extract_region(spectrum, region)
    else:
        calc_spectrum = spectrum

    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        flux = calc_spectrum.flux[~spectrum.mask]
        uncertainty = calc_spectrum.uncertainty.quantity[~spectrum.mask]
    else:
        flux = calc_spectrum.flux
        uncertainty = calc_spectrum.uncertainty.quantity

    # the axis=-1 will enable this to run on single-dispersion, single-flux
    # and single-dispersion, multiple-flux
    return np.mean(flux / uncertainty, axis=-1)


def snr_derived(spectrum, region=None):
    """
    This function computes the signal to noise ratio DER_SNR following the
    definition set forth by the Spectral Container Working Group of ST-ECF,
    MAST and CADC.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the equivalent width will be calculated.

    region: `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the SNR.

    Returns
    -------
    snr : `~astropy.units.Quantity` or list (based on region input)
        Signal to noise ratio of the spectrum or within the regions

    Notes
    -----
    The DER_SNR algorithm is an unbiased estimator describing the spectrum
    as a whole as long as the noise is uncorrelated in wavelength bins spaced
    two pixels apart, the noise is Normal distributed, for large wavelength
    regions, the signal over the scale of 5 or more pixels can be approximated
    by a straight line.

    The code and some documentation is derived from
    ``http://www.stecf.org/software/ASTROsoft/DER_SNR/der_snr.py``, and the
    algorithm itself is documented at https://esahubble.org/static/archives/stecfnewsletters/pdf/hst_stecf_0042.pdf
    """

    # No region, therefore whole spectrum.
    if region is None:
        return _snr_derived(spectrum)

    # Single region
    elif isinstance(region, SpectralRegion):
        return _snr_derived(spectrum, region=region)

    # List of regions
    elif isinstance(region, list):
        return [_snr_derived(spectrum, region=reg)
                for reg in region]


def _snr_derived(spectrum, region=None):
    """
    This function computes the signal to noise ratio DER_SNR following the
    definition set forth by the Spectral Container Working Group of ST-ECF,
    MAST and CADC

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the equivalent width will be calculated.

    region: `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the SNR.

    Returns
    -------
    snr : `~astropy.units.Quantity` or list (based on region input)
        Signal to noise ratio of the spectrum or within the regions

    Notes
    -----
    This is a helper function for the above `snr_derived()` method.

    """

    if region is not None:
        calc_spectrum = extract_region(spectrum, region)
    else:
        calc_spectrum = spectrum

    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        flux = calc_spectrum.flux[~calc_spectrum.mask]
    else:
        flux = calc_spectrum.flux

    # Values that are exactly zero (padded) are skipped
    n = len(flux)

    # For spectra shorter than this, no value can be returned
    if n > 4:
        signal = np.median(flux)
        noise  = 0.6052697 * np.median(np.abs(2.0 * flux[2:n-2] - flux[0:n-4] - flux[4:n]))
        return signal / noise
    else:
        return 0.0


def _convert_uncertainty(uncertainty, to_class):
    if isinstance(uncertainty, to_class):
        return uncertainty.quantity

    if isinstance(uncertainty, StdDevUncertainty):
        variance = uncertainty.quantity ** 2
    elif isinstance(uncertainty, VarianceUncertainty):
        variance = uncertainty.quantity
    elif isinstance(uncertainty, InverseVariance):
        variance = 1/uncertainty.quantity
    else:
        raise ValueError("uncertainties only supported in StdDev, Variance, or InverseVariance")

    if to_class == VarianceUncertainty:
        return variance
    elif to_class == InverseVariance:
        return 1/variance
    elif to_class == StdDevUncertainty:
        return variance**0.5
    else:
        raise ValueError("to_class only supports StdDev, Variance, or InverseVariance")



from astropy.nddata import VarianceUncertainty, StdDevUncertainty
import astropy.units as u

from .conftest import remote_access
from .. import SpectrumList
from ..io.default_loaders import dc_common as loaders

REMOTE_ID = "4059032"

GAMA_2QZ_TEST_FILENAME = "J113606.3+001155a.fit"
GAMA_2SLAQ_QSO_TEST_FILENAME = "J091726.21+003424.0_a14_040423.fit"
GAMA_GAMA_LT_TEST_FILENAME = "LTF_09_1128_0555.fit"
GAMA_GAMA_TEST_FILENAME = "G12_Y2_009_044.fit"
GAMA_MGC_TEST_FILENAME = "MGC23320.fit"
GAMA_WIGGLEZ_TEST_FILENAME = "Spectrum-195254.fit"
OZDES_TEST_FILENAME = "OzDES-DR2_04720.fits"
WIGGLEZ_TEST_FILENAME = "wig206635.fits"

OZDES_CONFIG = {
    "hdus": {
        "0": {"purpose": "combined_science"},
        "1": {"purpose": "combined_error_variance"},
        "2": {"purpose": "skip"},
        "cycle": {
            "0": {"purpose": "science"},
            "1": {"purpose": "error_variance"},
            "2": {"purpose": "skip"},
        },
    },
    "units": None,
    "wcs": None,
    "all_standard_units": True,
    "all_keywords": False,
    "valid_wcs": True,
}
GAMA_2QZ_CONFIG = {
    "hdus": None,
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CD1_1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": False,
}
GAMA_2SLAQ_QSO_CONFIG = {
    "hdus": None,
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": False,
}
GAMA_LT_CONFIG = {
    "hdus": {"0": {"purpose": "science"}, },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX",
        "pixel_reference_point_value_keyword": "CRVAL",
        "pixel_width_keyword": "CDELT",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}
GAMA_WIGGLEZ_CONFIG = {
    "hdus": {
        "0": {"purpose": "science"},
        "1": {"purpose": "error_variance"},
        "2": {"purpose": "skip"},
    },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CDELT1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}
GAMA_GAMA_CONFIG = {
    "hdu": {
        "1": {
            "purpose": "science",
            "units": {"flux_unit": "10^-17 erg/s/cm^2/A"},
        },
        "2": {"purpose": "error_stdev"},
        "3": {"purpose": "unreduced_science"},
        "4": {"purpose": "unreduced_error_stdev"},
        "5": {"purpose": "sky"},
    },
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CD1_1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": False,
    "valid_wcs": False,
}
GAMA_MGC_CONFIG = {
    "hdu": None,
    "wcs": {
        "pixel_reference_point_keyword": "CRPIX1",
        "pixel_reference_point_value_keyword": "CRVAL1",
        "pixel_width_keyword": "CD1_1",
        "wavelength_unit": "Angstrom",
    },
    "units": {"flux_unit": "count"},
    "all_standard_units": False,
    "all_keywords": True,
    "valid_wcs": False,
}


class TestSingleSplit:
    @remote_access([{'id': "4460981", 'filename': WIGGLEZ_TEST_FILENAME}])
    def test_wigglez(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="WiggleZ")

        assert len(spectra) == 1
        assert spectra[0].flux.unit == u.Unit("1e-16 erg / (A cm2 s)")
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)
        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': "4460981", 'filename': WIGGLEZ_TEST_FILENAME}])
    def test_wigglez_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)

        assert len(spectra) == 1
        assert spectra[0].flux.unit == u.Unit("1e-16 erg / (A cm2 s)")
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)
        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': OZDES_TEST_FILENAME}])
    def test_ozdes(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format=loaders.SINGLE_SPLIT_LABEL, **OZDES_CONFIG
        )

        # The test file has the combined obs, and 4 other sets
        assert len(spectra) == 5

        assert spectra[0].flux.unit == u.count / u.Angstrom
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)
        assert spectra[0].meta.get("label") is None
        assert spectra[0].meta.get("header") is not None

        assert spectra[1].flux.unit == u.count / u.Angstrom
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[1].uncertainty, VarianceUncertainty)
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None

        assert spectra[2].flux.unit == u.count / u.Angstrom
        assert spectra[2].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[2].uncertainty, VarianceUncertainty)
        assert spectra[2].meta.get("label") is not None
        assert spectra[2].meta.get("header") is not None

        assert spectra[3].flux.unit == u.count / u.Angstrom
        assert spectra[3].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[3].uncertainty, VarianceUncertainty)
        assert spectra[3].meta.get("label") is not None
        assert spectra[3].meta.get("header") is not None

        assert spectra[4].flux.unit == u.count / u.Angstrom
        assert spectra[4].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[4].uncertainty, VarianceUncertainty)
        assert spectra[4].meta.get("label") is not None
        assert spectra[4].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': OZDES_TEST_FILENAME}])
    def test_ozdes_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="OzDES")

        # The test file has the combined obs, and 4 other sets
        assert len(spectra) == 5

        assert spectra[0].flux.unit == u.count / u.Angstrom
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)
        assert spectra[0].meta.get("label") is None
        assert spectra[0].meta.get("header") is not None

        assert spectra[1].flux.unit == u.count / u.Angstrom
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[1].uncertainty, VarianceUncertainty)
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None

        assert spectra[2].flux.unit == u.count / u.Angstrom
        assert spectra[2].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[2].uncertainty, VarianceUncertainty)
        assert spectra[2].meta.get("label") is not None
        assert spectra[2].meta.get("header") is not None

        assert spectra[3].flux.unit == u.count / u.Angstrom
        assert spectra[3].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[3].uncertainty, VarianceUncertainty)
        assert spectra[3].meta.get("label") is not None
        assert spectra[3].meta.get("header") is not None

        assert spectra[4].flux.unit == u.count / u.Angstrom
        assert spectra[4].spectral_axis.unit == u.Angstrom
        assert isinstance(spectra[4].uncertainty, VarianceUncertainty)
        assert spectra[4].meta.get("label") is not None
        assert spectra[4].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_2QZ_TEST_FILENAME}])
    def test_gama_2qz(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format=loaders.SINGLE_SPLIT_LABEL,
            **GAMA_2QZ_CONFIG
        )
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_2QZ_TEST_FILENAME}])
    def test_gama_2qz_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="GAMA-2QZ")
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_2QZ_TEST_FILENAME}])
    def test_gama_2qz_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_2SLAQ_QSO_TEST_FILENAME}])
    def test_2slaq_qso(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format=loaders.SINGLE_SPLIT_LABEL,
            **GAMA_2SLAQ_QSO_CONFIG
        )
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_2SLAQ_QSO_TEST_FILENAME}])
    def test_2slaq_qso_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="GAMA-2SLAQ-QSO")
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_2SLAQ_QSO_TEST_FILENAME}])
    def test_2slaq_qso_normalised(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_GAMA_LT_TEST_FILENAME}])
    def test_gama_lt(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path,
            format=loaders.SINGLE_SPLIT_LABEL,
            **GAMA_LT_CONFIG
        )
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert spectra[0].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_GAMA_LT_TEST_FILENAME}])
    def test_gama_lt_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="GAMA-LT")
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert spectra[0].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_GAMA_LT_TEST_FILENAME}])
    def test_gama_lt_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert spectra[0].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_WIGGLEZ_TEST_FILENAME}])
    def test_gama_wigglez(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format=loaders.SINGLE_SPLIT_LABEL,
            **GAMA_WIGGLEZ_CONFIG
        )
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_WIGGLEZ_TEST_FILENAME}])
    def test_gama_wigglez_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="GAMA-WiggleZ")
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_WIGGLEZ_TEST_FILENAME}])
    def test_gama_wigglez_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 1

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, VarianceUncertainty)

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None


class TestMultilineSingle:
    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_GAMA_TEST_FILENAME}])
    def test_gama_gama(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format=loaders.MULTILINE_SINGLE_LABEL,
            **GAMA_GAMA_CONFIG
        )
        assert len(spectra) == 3

        assert spectra[0].flux.unit == u.Unit("10^-17 erg/s/cm^2/A")
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert spectra[2].flux.unit == u.count
        assert spectra[2].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
        assert isinstance(spectra[1].uncertainty, StdDevUncertainty)
        assert spectra[2].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None
        assert spectra[2].meta.get("label") is not None
        assert spectra[2].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_GAMA_TEST_FILENAME}])
    def test_gama_gama_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="GAMA")
        assert len(spectra) == 3

        assert spectra[0].flux.unit == u.Unit("10^-17 erg/s/cm^2/A")
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert spectra[2].flux.unit == u.count
        assert spectra[2].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
        assert isinstance(spectra[1].uncertainty, StdDevUncertainty)
        assert spectra[2].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None
        assert spectra[2].meta.get("label") is not None
        assert spectra[2].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_GAMA_TEST_FILENAME}])
    def test_gama_gama_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 3

        assert spectra[0].flux.unit == u.Unit("10^-17 erg/s/cm^2/A")
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert spectra[1].spectral_axis.unit == u.Angstrom
        assert spectra[2].flux.unit == u.count
        assert spectra[2].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
        assert isinstance(spectra[1].uncertainty, StdDevUncertainty)
        assert spectra[2].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None
        assert spectra[2].meta.get("label") is not None
        assert spectra[2].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_MGC_TEST_FILENAME}])
    def test_gama_mgc(self, remote_data_path):
        spectra = SpectrumList.read(
            remote_data_path, format=loaders.MULTILINE_SINGLE_LABEL,
            **GAMA_MGC_CONFIG
        )
        assert len(spectra) == 2

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert spectra[1].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
        assert spectra[1].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_MGC_TEST_FILENAME}])
    def test_gama_mgc_named_loader(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path, format="GAMA-MGC")
        assert len(spectra) == 2

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert spectra[1].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
        assert spectra[1].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None

    @remote_access([{'id': REMOTE_ID, 'filename': GAMA_MGC_TEST_FILENAME}])
    def test_gama_mgc_guess(self, remote_data_path):
        spectra = SpectrumList.read(remote_data_path)
        assert len(spectra) == 2

        assert spectra[0].flux.unit == u.count
        assert spectra[0].spectral_axis.unit == u.Angstrom
        assert spectra[1].flux.unit == u.count
        assert spectra[1].spectral_axis.unit == u.Angstrom

        assert isinstance(spectra[0].uncertainty, StdDevUncertainty)
        assert spectra[1].uncertainty is None

        assert spectra[0].meta.get("label") is not None
        assert spectra[0].meta.get("header") is not None
        assert spectra[1].meta.get("label") is not None
        assert spectra[1].meta.get("header") is not None



"""
Contains classes that serialize spectral data types into ASDF representations.
"""

from numpy.testing import assert_allclose
from astropy.units import allclose
import astropy.nddata
from asdf.yamlutil import (custom_tree_to_tagged_tree,
                           tagged_tree_to_custom_tree)

from ....spectra import Spectrum1D, SpectrumList
from ..types import SpecutilsType

__all__ = ['Spectrum1DType', 'SpectrumListType']

UNCERTAINTY_TYPE_MAPPING = {
    'std': astropy.nddata.StdDevUncertainty,
    'var': astropy.nddata.VarianceUncertainty,
    'ivar': astropy.nddata.InverseVariance,
    'unknown': astropy.nddata.UnknownUncertainty,
}


class Spectrum1DType(SpecutilsType):
    """
    ASDF tag implementation used to serialize/deserialize Spectrum1D objects
    """
    name = 'spectra/spectrum1d'
    types = [Spectrum1D]
    version = '1.0.0'

    @classmethod
    def to_tree(cls, obj, ctx):
        """
        Converts Spectrum1D object into tree used for YAML representation
        """
        node = {}
        node['flux'] = custom_tree_to_tagged_tree(obj.flux, ctx)
        node['spectral_axis'] = custom_tree_to_tagged_tree(obj.spectral_axis,
                                                           ctx)
        if obj.uncertainty is not None:
            node['uncertainty'] = {}
            node['uncertainty'][
                'uncertainty_type'] = obj.uncertainty.uncertainty_type
            data = custom_tree_to_tagged_tree(obj.uncertainty.array, ctx)
            node['uncertainty']['data'] = data

        return node

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts tree representation back into Spectrum1D object
        """
        flux = tagged_tree_to_custom_tree(tree['flux'], ctx)
        spectral_axis = tagged_tree_to_custom_tree(tree['spectral_axis'], ctx)
        uncertainty = tree.get('uncertainty', None)
        if uncertainty is not None:
            klass = UNCERTAINTY_TYPE_MAPPING[uncertainty['uncertainty_type']]
            data = tagged_tree_to_custom_tree(uncertainty['data'], ctx)
            uncertainty = klass(data)

        return Spectrum1D(flux=flux, spectral_axis=spectral_axis,
                          uncertainty=uncertainty)

    @classmethod
    def assert_equal(cls, old, new):
        """
        Equality method for use in ASDF unit tests
        """
        assert allclose(old.flux, new.flux)
        assert allclose(old.spectral_axis, new.spectral_axis)
        if old.uncertainty is None:
            assert new.uncertainty is None
        else:
            assert old.uncertainty.uncertainty_type == new.uncertainty.uncertainty_type
            assert_allclose(old.uncertainty.array, new.uncertainty.array)


class SpectrumListType(SpecutilsType):
    """
    ASDF tag implementation used to serialize/deserialize SpectrumList objects
    """
    name = 'spectra/spectrum_list'
    types = [SpectrumList]
    version = '1.0.0'

    @classmethod
    def to_tree(cls, obj, ctx):
        """
        Converts SpectrumList object into tree used for YAML representation
        """
        return [custom_tree_to_tagged_tree(spectrum, ctx) for spectrum in obj]

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts tree representation back into SpectrumList object
        """
        spectra = [tagged_tree_to_custom_tree(node, ctx) for node in tree]
        return SpectrumList(spectra)

    @classmethod
    def assert_equal(cls, old, new):
        """
        Equality test used in ASDF unit tests
        """
        assert len(old) == len(new)
        for x, y in zip(old, new):
            Spectrum1DType.assert_equal(x, y)



import astropy.units as u
from astropy.utils.decorators import lazyproperty
from astropy.coordinates import SpectralCoord
import numpy as np

__all__ = ['SpectralAxis']

# We don't want to run doctests in the docstrings we inherit from Quantity
__doctest_skip__ = ['SpectralAxis.*']


class SpectralAxis(SpectralCoord):
    """
    Coordinate object representing spectral values corresponding to a specific
    spectrum. Overloads SpectralCoord with additional information (currently
    only bin edges).

    Parameters
    ----------
    bin_specification: str, optional
        Must be "edges" or "centers". Determines whether specified axis values
        are interpreted as bin edges or bin centers. Defaults to "centers".
    """

    _equivalent_unit = SpectralCoord._equivalent_unit + (u.pixel,)

    def __new__(cls, value, *args, bin_specification="centers", **kwargs):

        # Convert to bin centers if bin edges were given, since SpectralCoord
        # only accepts centers
        if bin_specification == "edges":
            bin_edges = value
            value = SpectralAxis._centers_from_edges(value)

        obj = super().__new__(cls, value, *args, **kwargs)

        if bin_specification == "edges":
            obj._bin_edges = bin_edges

        return obj

    @staticmethod
    def _edges_from_centers(centers, unit):
        """
        Calculates interior bin edges based on the average of each pair of
        centers, with the two outer edges based on extrapolated centers added
        to the beginning and end of the spectral axis.
        """
        a = np.insert(centers, 0, 2*centers[0] - centers[1])
        b = np.append(centers, 2*centers[-1] - centers[-2])
        edges = (a + b) / 2
        return edges*unit

    @staticmethod
    def _centers_from_edges(edges):
        """
        Calculates the bin centers as the average of each pair of edges
        """
        return (edges[1:] + edges[:-1]) / 2

    @lazyproperty
    def bin_edges(self):
        """
        Calculates bin edges if the spectral axis was created with centers
        specified.
        """
        if hasattr(self, '_bin_edges'):
            return self._bin_edges
        else:
            return self._edges_from_centers(self.value, self.unit)

    def with_observer_stationary_relative_to(self, frame,
                                             velocity=None,
                                             preserve_observer_frame=False):
        if self.unit is u.pixel:
            raise u.UnitsError("Cannot transform spectral coordinates in pixel units")
        super().with_observer_stationary_relative_to(frame,
                                                     velocity=velocity,
                                                     preserve_observer_frame=preserve_observer_frame)

    def with_radial_velocity_shift(self, target_shift=None, observer_shift=None):
        if self.unit is u.pixel:
            raise u.UnitsError("Cannot transform spectral coordinates in pixel units")
        return super().with_radial_velocity_shift(target_shift=target_shift,
                                                  observer_shift=observer_shift)



"""
A module for analysis tools focused on determining the width of
spectral features.
"""

import numpy as np
from astropy.nddata import StdDevUncertainty
from astropy.stats.funcs import gaussian_sigma_to_fwhm
from ..manipulation import extract_region
from . import centroid
from .uncertainty import _convert_uncertainty
from .utils import computation_wrapper
from scipy.signal import find_peaks, peak_widths


__all__ = ['gaussian_sigma_width', 'gaussian_fwhm', 'fwhm', 'fwzi']


def gaussian_sigma_width(spectrum, regions=None):
    """
    Estimate the width of the spectrum using a second-moment analysis.

    The value is scaled to match the sigma/standard deviation parameter of a
    standard Gaussian profile. This will be calculated over the regions, if
    they are specified.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object over which the width will be calculated.

    regions: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the gaussian sigma width. If
        regions is `None`, computation is performed over entire spectrum.

    Returns
    -------
    approx_sigma: `~astropy.units.Quantity` or list (based on region input)
        Approximated sigma value of the spectrum

    Notes
    -----
    The spectrum should be continuum subtracted before being passed to this
    function.
    """
    return computation_wrapper(_compute_gaussian_sigma_width, spectrum, regions)


def gaussian_fwhm(spectrum, regions=None):
    """
    Estimate the width of the spectrum using a second-moment analysis.

    The value is scaled to match the full width at half max of a standard
    Gaussian profile.  This will be calculated over the regions, if they are
    specified.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object overwhich the width will be calculated.

    regions: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the FWHM value. If regions is
        `None`, computation is performed over entire spectrum.

    Returns
    -------
    gaussian_fwhm : `~astropy.units.Quantity` or list (based on region input)
        Approximate full width of the signal at half max

    Notes
    -----
    The spectrum should be continuum subtracted before being passed to this
    function.
    """
    return computation_wrapper(_compute_gaussian_fwhm, spectrum, regions)


def fwhm(spectrum, regions=None):
    """
    Compute the true full width half max of the spectrum.

    This makes no assumptions about the shape of the spectrum (e.g. whether it
    is Gaussian). It finds the maximum of the spectrum, and then locates the
    point closest to half max on either side of the maximum, and
    measures the distance between them. This will be calculated over the
    regions, if they are specified.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object over which the width will be calculated.

    regions: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the FWHM value. If regions is
        `None`, computation is performed over entire spectrum.

    Returns
    -------
    whm : `~astropy.units.Quantity` or list (based on region input)
        Full width of the signal at half max

    Notes
    -----
    The spectrum should be continuum subtracted before being passed to this
    function.
    """
    return computation_wrapper(_compute_fwhm, spectrum, regions)


def fwzi(spectrum, regions=None):
    """
    Compute the true full width at zero intensity (i.e. the continuum level)
    of the spectrum.

    This makes no assumptions about the shape of the spectrum. It uses the
    scipy peak-finding to determine the index of the highest flux value, and
    then calculates width at that base of the feature.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object over which the width will be calculated.

    regions: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the FWZI value. If regions is
        `None`, computation is performed over entire spectrum.

    Returns
    -------
    `~astropy.units.Quantity` or list (based on region input)
        Full width of the signal at zero intensity.

    Notes
    -----
    The spectrum must be continuum subtracted before being passed to this
    function.
    """
    return computation_wrapper(_compute_fwzi, spectrum, regions)


def _compute_fwzi(spectrum, regions=None):
    if regions is not None:
        calc_spectrum = extract_region(spectrum, regions)
    else:
        calc_spectrum = spectrum

    # Create a copy of the flux array to ensure the value on the spectrum
    # object is not altered.
    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        disp = calc_spectrum.spectral_axis[~spectrum.mask]
        flux = calc_spectrum.flux[~spectrum.mask].copy()
    else:
        disp = calc_spectrum.spectral_axis
        flux = calc_spectrum.flux.copy()

    # For noisy data, ensure that the search from the centroid stops on
    # either side once the flux value reaches zero.
    flux[flux < 0] = 0

    def find_width(data):
        # Find the peaks in the flux data
        peaks, _ = find_peaks(data)

        # Find the index of the maximum peak value in the found peak list
        peak_ind = [peaks[np.argmin(np.abs(
            np.array(peaks) - np.argmin(np.abs(data - np.max(data)))))]]

        # Calculate the width for the given feature
        widths, _, _, _ = \
            peak_widths(data, peak_ind, rel_height=1-1e-7)

        return widths[0] * disp.unit

    if flux.ndim > 1:
        tot_widths = []

        for i in range(flux.shape[0]):
            tot_widths.append(find_width(flux[i]))

        return tot_widths

    return find_width(flux)


def _compute_gaussian_fwhm(spectrum, regions=None):
    """
    This is a helper function for the above `gaussian_fwhm()` method.
    """

    sigma = _compute_gaussian_sigma_width(spectrum, regions)

    fwhm = sigma * gaussian_sigma_to_fwhm
    if sigma.uncertainty is not None:
        fwhm.uncertainty = sigma.uncertainty * gaussian_sigma_to_fwhm
    else:
        fwhm.uncertainty = None

    fwhm.uncertainty_type = 'std'
    return fwhm


def _compute_gaussian_sigma_width(spectrum, regions=None):
    """
    This is a helper function for the above `gaussian_sigma_width()` method.
    """

    if regions is not None:
        calc_spectrum = extract_region(spectrum, regions)
    else:
        calc_spectrum = spectrum

    if spectrum.uncertainty is not None:
        flux_uncert = _convert_uncertainty(calc_spectrum.uncertainty, StdDevUncertainty)
    else:
        # dummy value for uncertainties to avoid extra if-statements when applying mask
        flux_uncert = np.zeros_like(calc_spectrum.flux)

    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        flux = calc_spectrum.flux[~spectrum.mask]
        spectral_axis = calc_spectrum.spectral_axis[~spectrum.mask]
        flux_uncert = flux_uncert[~calc_spectrum.mask]
    else:
        flux = calc_spectrum.flux
        spectral_axis = calc_spectrum.spectral_axis

    centroid_result = centroid(spectrum, regions)

    if flux.ndim > 1:
        spectral_axis = np.broadcast_to(spectral_axis, flux.shape, subok=True)
        centroid_result_uncert = centroid_result.uncertainty
        centroid_result = centroid_result[:, np.newaxis]
        centroid_result.uncertainty = centroid_result_uncert[:, np.newaxis] if centroid_result_uncert is not None else None  # noqa

    dx = (spectral_axis - centroid_result)
    numerator = np.sum((dx * dx) * flux, axis=-1)
    denom = np.sum(flux, axis=-1)
    sigma2 = numerator / denom
    sigma = np.sqrt(sigma2)
    if centroid_result.uncertainty is not None:
        # NOTE: until/unless disp_uncert is supported, dx_uncert == centroid_result.uncertainty
        disp_uncert = 0.0 * spectral_axis.unit
        dx_uncert = np.sqrt(disp_uncert**2 + centroid_result.uncertainty**2)
        # dx_uncert = centroid_result.uncertainty

        # uncertainty for each term in the numerator sum
        num_term_uncerts = dx * dx * flux * np.sqrt(2*(dx_uncert/dx)**2 + (flux_uncert/flux)**2)
        # uncertainty (squared) for the numerator, added in quadrature
        num_uncertsq = np.sum(num_term_uncerts**2, axis=-1)
        # uncertainty (squared) for the denomenator
        denom_uncertsq = np.sum(flux_uncert**2)

        sigma2_uncert = numerator/denom * np.sqrt(num_uncertsq * numerator**-2 +
                                                  denom_uncertsq * denom**-2)

        sigma.uncertainty = 0.5 * sigma2_uncert / sigma2 * sigma.unit
    else:
        sigma.uncertainty = None

    sigma.uncertainty_type = 'std'
    return sigma


def _compute_single_fwhm(flux, spectral_axis):
    """
    This is a helper function for the above `fwhm()` method.
    """

    # The .value attribute is used here as the following algorithm does not
    # use any array operations and would otherwise introduce a relatively
    # significant overhead factor.  Two-point linear interpolation is used to
    # achieve sub-pixel precision.
    flux_value = flux.value
    spectral_value = spectral_axis.value

    argmax = flux_value.argmax()
    halfval = flux_value[argmax] / 2
    left = flux_value[:argmax] < halfval
    right = flux_value[argmax + 1:] < halfval

    # Highest signal at the first point
    i0 = np.nonzero(left)[0]
    if i0.size == 0:
        left_value = spectral_value[0]
    else:
        i0 = i0[-1]
        i1 = i0 + 1
        left_flux = flux_value[i0]
        left_spectral = spectral_value[i0]
        left_value = ((halfval - left_flux)
                      * (spectral_value[i1] - left_spectral)
                      / (flux_value[i1] - left_flux)
                      + left_spectral)

    # Highest signal at the last point
    i1 = np.nonzero(right)[0]
    if i1.size == 0:
        right_value = spectral_value[-1]
    else:
        i1 = i1[0] + argmax + 1
        i0 = i1 - 1
        left_flux = flux_value[i0]
        left_spectral = spectral_value[i0]
        right_value = ((halfval - left_flux)
                       * (spectral_value[i1] - left_spectral)
                       / (flux_value[i1] - left_flux)
                       + left_spectral)

    return spectral_axis.unit * np.abs(right_value - left_value)


def _compute_fwhm(spectrum, regions=None):
    """
    This is a helper function for the above `fwhm()` method.
    """

    if regions is not None:
        calc_spectrum = extract_region(spectrum, regions)
    else:
        calc_spectrum = spectrum

    flux = calc_spectrum.flux
    spectral_axis = calc_spectrum.spectral_axis

    if flux.ndim > 1:
        return [_compute_single_fwhm(x, spectral_axis) for x in flux]
    else:
        return _compute_single_fwhm(flux, spectral_axis)



import astropy.units as u
import numpy as np
import pytest
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning

from ..analysis import centroid, fwhm
from ..fitting import (estimate_line_parameters, find_lines_derivative,
                       find_lines_threshold, fit_lines)
from ..manipulation import (extract_region, noise_region_uncertainty,
                            spectrum_from_model)
from ..spectra import SpectralRegion, Spectrum1D


def single_peak():
    np.random.seed(0)
    x = np.linspace(0., 10., 200)
    y_single = 3 * np.exp(-0.5 * (x - 6.3)**2 / 0.8**2)
    y_single += np.random.normal(0., 0.2, x.shape)
    return x, y_single


def single_peak_continuum():
    np.random.seed(0)
    x = np.linspace(0., 10., 200)
    y_single = 3 * np.exp(-0.5 * (x - 6.3)**2 / 0.3**2)
    y_single += np.random.normal(0., 0.2, x.shape)

    y_continuum = 3.2 * np.exp(-0.5 * (x - 0.6)**2 / 2.8**2)
    y_single += y_continuum
    return x, y_single


def single_peak_extra():
    x, y_single = single_peak()
    extra = 4 * np.exp(-0.5 * (x + 8.3)**2 / 0.1**2)
    y_single_extra = y_single + extra
    return x, y_single_extra


def double_peak():
    np.random.seed(42)
    g1 = models.Gaussian1D(1, 4.6, 0.2)
    g2 = models.Gaussian1D(2.5, 5.5, 0.1)
    x = np.linspace(0, 10, 200)
    y_double = g1(x) + g2(x) + np.random.normal(0., 0.2, x.shape)
    return x, y_double


def double_peak_absorption_and_emission():
    np.random.seed(42)
    g1 = models.Gaussian1D(1, 4.6, 0.2)
    g2 = models.Gaussian1D(2.5, 5.5, 0.1)
    g3 = models.Gaussian1D(-1.7, 8.2, 0.1)
    x = np.linspace(0, 10, 200)
    y_double = g1(x) + g2(x) + g3(x) + np.random.normal(0., 0.2, x.shape)
    return x, y_double


def test_find_lines_derivative():

    # Create the spectrum to fit
    x_double, y_double = double_peak_absorption_and_emission()
    spectrum = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um)

    # Derivative method
    with pytest.warns(AstropyUserWarning,
                      match='Spectrum is not below the threshold signal-to-noise'):
        lines = find_lines_derivative(spectrum, flux_threshold=0.75)

    emission_lines = lines[lines['line_type'] == 'emission']
    absorption_lines = lines[lines['line_type'] == 'absorption']

    assert emission_lines['line_center_index'].tolist() == [90, 109]
    assert absorption_lines['line_center_index'].tolist() == [163]


def test_find_lines_threshold():

    # Create the spectrum to fit
    x_double, y_double = double_peak_absorption_and_emission()
    spectrum = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um)

    # Derivative method
    noise_region = SpectralRegion(0*u.um, 3*u.um)
    spectrum = noise_region_uncertainty(spectrum, noise_region)
    with pytest.warns(AstropyUserWarning,
                      match='Spectrum is not below the threshold signal-to-noise'):
        lines = find_lines_threshold(spectrum, noise_factor=3)

    emission_lines = lines[lines['line_type'] == 'emission']
    absorption_lines = lines[lines['line_type'] == 'absorption']

    assert emission_lines['line_center_index'].tolist() == [91, 96, 109, 179]
    assert absorption_lines['line_center_index'].tolist() == [163]


def test_single_peak_estimate():
    """
    Single Peak fit.
    """

    # Create the spectrum
    x_single, y_single = single_peak()
    s_single = Spectrum1D(flux=y_single*u.Jy, spectral_axis=x_single*u.um)

    #
    # Estimate parameter Gaussian1D
    # we give the true values for the Gaussian because it actually *should*
    # be pretty close to the true values, because it's a Gaussian...
    #

    g_init = estimate_line_parameters(s_single, models.Gaussian1D())

    assert np.isclose(g_init.amplitude.value, 3., rtol=.2)
    assert np.isclose(g_init.mean.value, 6.3, rtol=.1)
    assert np.isclose(g_init.stddev.value, 0.8, rtol=.3)

    assert g_init.amplitude.unit == u.Jy
    assert g_init.mean.unit == u.um
    assert g_init.stddev.unit == u.um

    #
    # Estimate parameter Lorentz1D
    # unlike the Gaussian1D here we do hand-picked comparison values, because
    # the "single peak" is a Gaussian and therefore the Lorentzian fit shouldn't
    # be quite right anyway
    #

    g_init = estimate_line_parameters(s_single, models.Lorentz1D())

    assert np.isclose(g_init.amplitude.value, 3.354169257846847)
    assert np.isclose(g_init.x_0.value, 6.218588636687762)
    assert np.isclose(g_init.fwhm.value, 1.6339001193853715)

    assert g_init.amplitude.unit == u.Jy
    assert g_init.x_0.unit == u.um
    assert g_init.fwhm.unit == u.um

    #
    # Estimate parameter Voigt1D
    #

    g_init = estimate_line_parameters(s_single, models.Voigt1D())

    assert np.isclose(g_init.amplitude_L.value, 3.354169257846847)
    assert np.isclose(g_init.x_0.value, 6.218588636687762)
    assert np.isclose(g_init.fwhm_L.value, 1.1553418541989058)
    assert np.isclose(g_init.fwhm_G.value, 1.1553418541989058)

    assert g_init.amplitude_L.unit == u.Jy
    assert g_init.x_0.unit == u.um
    assert g_init.fwhm_L.unit == u.um
    assert g_init.fwhm_G.unit == u.um

    #
    # Estimate parameter RickerWavelet1D
    #
    mh = models.RickerWavelet1D
    mh.amplitude.estimator = lambda s: max(s.flux)
    mh.x_0.estimator = lambda s, region: centroid(s, regions=region)
    mh.sigma.estimator = lambda s, regions: fwhm(s, regions)

    g_init = estimate_line_parameters(s_single, mh)

    assert np.isclose(g_init.amplitude.value, 3.354169257846847)
    assert np.isclose(g_init.x_0.value, 6.218588636687762)
    assert np.isclose(g_init.sigma.value, 1.6339001193853715)

    assert g_init.amplitude.unit == u.Jy
    assert g_init.x_0.unit == u.um
    assert g_init.sigma.unit == u.um


def test_single_peak_fit():
    """
    Single peak fit
    """

    # Create the spectrum
    x_single, y_single = single_peak()
    s_single = Spectrum1D(flux=y_single*u.Jy, spectral_axis=x_single*u.um)

    # Fit the spectrum
    g_init = models.Gaussian1D(amplitude=3.*u.Jy, mean=6.1*u.um, stddev=1.*u.um)
    g_fit = fit_lines(s_single, g_init)
    y_single_fit = g_fit(x_single*u.um)

    # Comparing every 10th value.
    y_single_fit_expected = np.array([3.69669474e-13, 3.57992454e-11, 2.36719426e-09, 1.06879318e-07,
               3.29498310e-06, 6.93605383e-05, 9.96945607e-04, 9.78431032e-03,
               6.55675141e-02, 3.00017760e-01, 9.37356842e-01, 1.99969007e+00,
               2.91286375e+00, 2.89719280e+00, 1.96758892e+00, 9.12412206e-01,
               2.88900005e-01, 6.24602556e-02, 9.22061121e-03, 9.29427266e-04]) * u.Jy

    assert np.allclose(y_single_fit.value[::10], y_single_fit_expected.value, atol=1e-5)
    assert np.allclose(list(g_fit.stds), [0.04627, 0.01427, 0.01427], rtol=0.05)


def test_single_peak_fit_with_uncertainties():
    """
    Single peak fit
    """

    # Create the spectrum
    line_mod = models.Gaussian1D(amplitude=100,
                                 mean=6563*u.angstrom,
                                 stddev=20*u.angstrom) + models.Const1D(10)

    init_mod = models.Gaussian1D(amplitude=85 * u.Jy,
                                 mean=6550*u.angstrom,
                                 stddev=30*u.angstrom) + models.Const1D(8 * u.Jy)
    x = np.linspace(6400, 6700, 300) * u.AA

    def calculate_rms(x, init_mod, implicit_weights):
        rms = []

        for _ in range(100):
            ymod = line_mod(x)
            y = np.random.poisson(ymod)
            unc = np.sqrt(ymod)

            spec = Spectrum1D(spectral_axis=x, flux=y * u.Jy,
                              uncertainty=StdDevUncertainty(unc * u.Jy))

            weights = 'unc' if implicit_weights else unc ** -1

            spec_fit = fit_lines(spec, init_mod, weights=weights)

            rms.append(np.std(spec_fit(x).value - y))

        return np.median(rms)

    assert np.allclose(calculate_rms(x, init_mod, implicit_weights=True),
                       5.101611033799086)
    assert np.allclose(calculate_rms(x, init_mod, implicit_weights=False),
                       5.113697654869089)


def test_single_peak_fit_window():
    """
    Single Peak fit with a window specified
    """

    # Create the sepctrum
    x_single, y_single = single_peak()
    s_single = Spectrum1D(flux=y_single*u.Jy, spectral_axis=x_single*u.um)

    # Fit the spectrum
    g_init = models.Gaussian1D(amplitude=3.*u.Jy, mean=5.5*u.um, stddev=1.*u.um)
    g_fit = fit_lines(s_single, g_init, window=2*u.um)
    y_single_fit = g_fit(x_single*u.um)

    # Comparing every 10th value.
    y_single_fit_expected = np.array([3.69669474e-13, 3.57992454e-11, 2.36719426e-09, 1.06879318e-07,
                                      3.29498310e-06, 6.93605383e-05, 9.96945607e-04, 9.78431032e-03,
                                      6.55675141e-02, 3.00017760e-01, 9.37356842e-01, 1.99969007e+00,
                                      2.91286375e+00, 2.89719280e+00, 1.96758892e+00, 9.12412206e-01,
                                      2.88900005e-01, 6.24602556e-02, 9.22061121e-03, 9.29427266e-04]) * u.Jy

    assert np.allclose(y_single_fit.value[::10], y_single_fit_expected.value, atol=1e-5)


def test_single_peak_fit_tuple_window():
    """
    Single Peak fit with a window specified as a tuple
    """

    # Create the spectrum to fit
    x_single, y_single = single_peak()
    s_single = Spectrum1D(flux=y_single*u.Jy, spectral_axis=x_single*u.um)

    # Fit the spectrum
    g_init = models.Gaussian1D(amplitude=3.*u.Jy, mean=5.5*u.um, stddev=1.*u.um)
    g_fit = fit_lines(s_single, g_init, window=(6*u.um, 7*u.um))
    y_single_fit = g_fit(x_single*u.um)

    # Comparing every 10th value.
    y_single_fit_expected = np.array([2.29674788e-16, 6.65518998e-14, 1.20595958e-11, 1.36656472e-09,
                                      9.68395624e-08, 4.29141576e-06, 1.18925100e-04, 2.06096976e-03,
                                      2.23354585e-02, 1.51371211e-01, 6.41529836e-01, 1.70026100e+00,
                                      2.81799025e+00, 2.92071068e+00, 1.89305291e+00, 7.67294570e-01,
                                      1.94485245e-01, 3.08273612e-02, 3.05570344e-03, 1.89413625e-04])*u.Jy

    assert np.allclose(y_single_fit.value[::10], y_single_fit_expected.value, atol=1e-5)


def test_double_peak_fit():
    """
    Double Peak fit.
    """

    # Create the spectrum to fit
    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um)

    # Fit the spectrum
    g1_init = models.Gaussian1D(amplitude=2.3*u.Jy, mean=5.6*u.um, stddev=0.1*u.um)
    g2_init = models.Gaussian1D(amplitude=1.*u.Jy, mean=4.4*u.um, stddev=0.1*u.um)
    g12_fit = fit_lines(s_double, g1_init+g2_init)
    y12_double_fit = g12_fit(x_double*u.um)

    # Comparing every 10th value.
    y12_double_fit_expected = np.array([2.86790780e-130, 2.12984643e-103, 1.20060032e-079, 5.13707226e-059,
                                        1.66839912e-041, 4.11292970e-027, 7.69608184e-016, 1.09308800e-007,
                                        1.17844042e-002, 9.64333366e-001, 6.04322205e-002, 2.22653307e+000,
                                        5.51964567e-005, 8.13581859e-018, 6.37320251e-038, 8.85834856e-055,
                                        1.05230522e-074, 9.48850399e-098, 6.49412764e-124, 3.37373489e-153])

    assert np.allclose(y12_double_fit.value[::10], y12_double_fit_expected, atol=1e-5)


def test_double_peak_fit_tuple_window():
    """
    Doulbe Peak fit with a window specified as a tuple
    """

    # Create the spectrum to fit
    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um, rest_value=0*u.um)

    # Fit the spectrum.
    g2_init = models.Gaussian1D(amplitude=1.*u.Jy, mean=4.7*u.um, stddev=0.2*u.um)
    g2_fit = fit_lines(s_double, g2_init, window=(4.3*u.um, 5.3*u.um))
    y2_double_fit = g2_fit(x_double*u.um)

    # Comparing every 10th value.
    y2_double_fit_expected = np.array([2.82386634e-116, 2.84746284e-092, 4.63895634e-071, 1.22104254e-052,
                                       5.19265653e-037, 3.56776869e-024, 3.96051875e-014, 7.10322789e-007,
                                       2.05829545e-002, 9.63624806e-001, 7.28880815e-002, 8.90744929e-006,
                                       1.75872724e-012, 5.61037526e-022, 2.89156942e-034, 2.40781783e-049,
                                       3.23938019e-067, 7.04122962e-088, 2.47276807e-111, 1.40302869e-137])

    assert np.allclose(y2_double_fit.value[::10], y2_double_fit_expected, atol=1e-5)


def test_double_peak_fit_window():
    """
    Double Peak fit with a window.
    """

    # Create the specturm to fit
    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um, rest_value=0*u.um)

    # Fit the spectrum
    g2_init = models.Gaussian1D(amplitude=1.*u.Jy, mean=4.7*u.um, stddev=0.2*u.um)
    g2_fit = fit_lines(s_double, g2_init, window=0.3*u.um)
    y2_double_fit = g2_fit(x_double*u.um)

    # Comparing every 10th value.
    y2_double_fit_expected = np.array([1.66363393e-128, 5.28910721e-102, 1.40949521e-078, 3.14848385e-058,
                                       5.89516506e-041, 9.25224449e-027, 1.21718016e-015, 1.34220626e-007,
                                       1.24062432e-002, 9.61209273e-001, 6.24240938e-002, 3.39815491e-006,
                                       1.55056770e-013, 5.93054936e-024, 1.90132233e-037, 5.10943886e-054,
                                       1.15092572e-073, 2.17309153e-096, 3.43926290e-122, 4.56256813e-151])

    assert np.allclose(y2_double_fit.value[::10], y2_double_fit_expected, atol=1e-5)


def test_double_peak_fit_separate_window():
    """
    Double Peak fit with a window.
    """

    # Create the spectrum to fit
    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um, rest_value=0*u.um)

    # Fit the spectrum
    gl_init = models.Gaussian1D(amplitude=1.*u.Jy, mean=4.8*u.um, stddev=0.2*u.um)
    gr_init = models.Gaussian1D(amplitude=2.*u.Jy, mean=5.3*u.um, stddev=0.2*u.um)
    gl_fit, gr_fit = fit_lines(s_double, [gl_init, gr_init], window=0.2*u.um)
    yl_double_fit = gl_fit(x_double*u.um)
    yr_double_fit = gr_fit(x_double*u.um)

    # Comparing every 10th value.
    yl_double_fit_expected = np.array([3.40725147e-18, 5.05500395e-15, 3.59471319e-12, 1.22527176e-09,
                                       2.00182467e-07, 1.56763547e-05, 5.88422893e-04, 1.05866724e-02,
                                       9.12966452e-02, 3.77377148e-01, 7.47690410e-01, 7.10057397e-01,
                                       3.23214276e-01, 7.05201207e-02, 7.37498248e-03, 3.69687164e-04,
                                       8.88245844e-06, 1.02295712e-07, 5.64686114e-10, 1.49410879e-12])

    assert np.allclose(yl_double_fit.value[::10], yl_double_fit_expected, atol=1e-5)

    # Comparing every 10th value.
    yr_double_fit_expected = np.array([0.00000000e+000, 0.00000000e+000, 0.00000000e+000, 3.04416285e-259,
                                       3.85323221e-198, 2.98888589e-145, 1.42075875e-100, 4.13864520e-064,
                                       7.38793226e-036, 8.08191847e-016, 5.41792361e-004, 2.22575901e+000,
                                       5.60338234e-005, 8.64468603e-018, 8.17287853e-039, 4.73508430e-068,
                                       1.68115300e-105, 3.65774659e-151, 4.87693358e-205, 3.98480359e-267])

    assert np.allclose(yr_double_fit.value[::10], yr_double_fit_expected, atol=1e-5)


def test_double_peak_fit_separate_window_tuple_window():
    """
    Double Peak fit with a window.
    """

    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um, rest_value=0*u.um)

    g1_init = models.Gaussian1D(amplitude=2.*u.Jy, mean=5.3*u.um, stddev=0.2*u.um)
    g2_init = models.Gaussian1D(amplitude=1.*u.Jy, mean=4.9*u.um, stddev=0.1*u.um)
    g1_fit, g2_fit = fit_lines(s_double, [g1_init, g2_init], window=[(5.3*u.um, 5.8*u.um), (4.6*u.um, 5.3*u.um)])
    y1_double_fit = g1_fit(x_double*u.um)
    y2_double_fit = g2_fit(x_double*u.um)

    # Comparing every 10th value.
    y1_double_fit_expected = np.array([0.00000000e+000, 0.00000000e+000, 5.61595149e-307, 3.38362505e-242,
                                       4.27358433e-185, 1.13149721e-135, 6.28008984e-094, 7.30683649e-060,
                                       1.78214929e-033, 9.11192086e-015, 9.76623021e-004, 2.19429562e+000,
                                       1.03350951e-004, 1.02043415e-016, 2.11206194e-036, 9.16388177e-064,
                                       8.33495900e-099, 1.58920023e-141, 6.35191874e-192, 5.32209240e-250])

    assert np.allclose(y1_double_fit.value[::10], y1_double_fit_expected, atol=1e-5)

    # Comparing every 10th value.
    y2_double_fit_expected = np.array([2.52990802e-158, 5.15446435e-126, 2.07577138e-097, 1.65231432e-072,
                                       2.59969849e-051, 8.08482210e-034, 4.96975664e-020, 6.03833143e-010,
                                       1.45016006e-003, 6.88386116e-001, 6.45900222e-002, 1.19788723e-006,
                                       4.39120391e-015, 3.18176751e-027, 4.55691000e-043, 1.28999976e-062,
                                       7.21815119e-086, 7.98324559e-113, 1.74521997e-143, 7.54115780e-178])

    assert np.allclose(y2_double_fit.value[::10], y2_double_fit_expected, atol=1e-3)


def test_double_peak_fit_with_exclusion():
    """
    Double Peak fit with a window.
    """

    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um, rest_value=0*u.um)

    g1_init = models.Gaussian1D(amplitude=1.*u.Jy, mean=4.9*u.um, stddev=0.2*u.um)
    g1_fit = fit_lines(s_double, g1_init, exclude_regions=[SpectralRegion(5.2*u.um, 5.8*u.um)])
    y1_double_fit = g1_fit(x_double*u.um)

    # Comparing every 10th value.
    y1_double_fit_expected = np.array([1.22845792e-129, 6.72991033e-103, 2.89997996e-079, 9.82918347e-059,
                                       2.62045444e-041, 5.49506092e-027, 9.06369193e-016, 1.17591109e-007,
                                       1.19999812e-002, 9.63215290e-001, 6.08139060e-002, 3.02008492e-006,
                                       1.17970152e-013, 3.62461471e-024, 8.75968088e-038, 1.66514266e-054,
                                       2.48972475e-074, 2.92811381e-097, 2.70870241e-123, 1.97093074e-152])

    assert np.allclose(y1_double_fit.value[::10], y1_double_fit_expected, atol=1e-5)


def tie_center(model):
    """ Dummy method for testing passing of tied parameter """
    mean = 50 * model.stddev
    return mean


def test_fixed_parameters():
    """
    Test to confirm fixed parameters do not change.
    """

    x = np.linspace(0., 10., 200)
    y = 3 * np.exp(-0.5 * (x - 6.3)**2 / 0.8**2)
    y += np.random.normal(0., 0.2, x.shape)
    spectrum = Spectrum1D(flux=y*u.Jy, spectral_axis=x*u.um)

    # Test passing fixed and bounds parameters
    g_init = models.Gaussian1D(amplitude=3.*u.Jy, mean=6.1*u.um, stddev=1.*u.um,
                               fixed={'mean': True},
                               bounds={'amplitude': (2, 5)*u.Jy},
                               name="Gaussian Test Model")

    g_fit = fit_lines(spectrum, g_init)

    assert_quantity_allclose(g_fit.mean, 6.1*u.um)
    assert g_fit.bounds == g_init.bounds
    assert g_fit.name == "Gaussian Test Model"

    # Test passing of tied parameter
    g_init = models.Gaussian1D(amplitude=3.*u.Jy, mean=6.1*u.um, stddev=1.*u.um)
    g_init.mean.tied = tie_center
    g_fit = fit_lines(spectrum, g_init)

    assert g_fit.tied == g_init.tied
    assert g_fit.name == g_init.name


def test_name_preservation_after_fitting():
    """
    Test to confirm model and submodels names are preserved after fitting.
    """

    x = np.linspace(0., 10., 200)
    y = 3 * np.exp(-0.5 * (x - 6.3)**2 / 0.8**2)
    y += np.random.normal(0., 0.2, x.shape)
    spectrum = Spectrum1D(flux=y*u.Jy, spectral_axis=x*u.um)

    subcomponents = models.Gaussian1D(name="Model I") * models.Gaussian1D(name="Second Model")
    c_model = subcomponents + models.Gaussian1D(name="Model 3")
    c_model.name = "Compound Model with 3 components"
    model_fit = fit_lines(spectrum, c_model)

    assert model_fit.name == "Compound Model with 3 components"
    assert model_fit.submodel_names == ("Model I", "Second Model", "Model 3")


def test_ignore_units():
    """
    Ignore the units
    """

    #
    #  Ignore the units based on there not being units on the model
    #

    # Create the spectrum
    x_single, y_single = single_peak()
    s_single = Spectrum1D(flux=y_single*u.Jy, spectral_axis=x_single*u.um)

    # Fit the spectrum
    g_init = models.Gaussian1D(amplitude=3, mean=6.1, stddev=1.)
    g_fit = fit_lines(s_single, g_init)
    y_single_fit = g_fit(x_single*u.um)

    # Comparing every 10th value.
    y_single_fit_expected = np.array([3.69669474e-13, 3.57992454e-11, 2.36719426e-09, 1.06879318e-07,
                                      3.29498310e-06, 6.93605383e-05, 9.96945607e-04, 9.78431032e-03,
                                      6.55675141e-02, 3.00017760e-01, 9.37356842e-01, 1.99969007e+00,
                                      2.91286375e+00, 2.89719280e+00, 1.96758892e+00, 9.12412206e-01,
                                      2.88900005e-01, 6.24602556e-02, 9.22061121e-03, 9.29427266e-04])

    assert np.allclose(y_single_fit.value[::10], y_single_fit_expected, atol=1e-5)
    assert y_single_fit.unit == s_single.flux.unit

    #
    # Ignore the units based on not being in the model
    #

    # Create the spectrum to fit
    x_double, y_double = double_peak()
    s_double = Spectrum1D(flux=y_double*u.Jy, spectral_axis=x_double*u.um)

    # Fit the spectrum
    g1_init = models.Gaussian1D(amplitude=2.3, mean=5.6, stddev=0.1)
    g2_init = models.Gaussian1D(amplitude=1., mean=4.4, stddev=0.1)
    g12_fit = fit_lines(s_double, g1_init+g2_init)
    y12_double_fit = g12_fit(x_double*u.um)

    # Comparing every 10th value.
    y12_double_fit_expected = np.array([2.86790780e-130, 2.12984643e-103, 1.20060032e-079, 5.13707226e-059,
                                        1.66839912e-041, 4.11292970e-027, 7.69608184e-016, 1.09308800e-007,
                                        1.17844042e-002, 9.64333366e-001, 6.04322205e-002, 2.22653307e+000,
                                        5.51964567e-005, 8.13581859e-018, 6.37320251e-038, 8.85834856e-055,
                                        1.05230522e-074, 9.48850399e-098, 6.49412764e-124, 3.37373489e-153])

    assert np.allclose(y12_double_fit.value[::10], y12_double_fit_expected, atol=1e-5)


def test_fitter_parameters():
    """
    Single Peak fit.
    """

    # Create the spectrum
    x_single, y_single = single_peak()
    s_single = Spectrum1D(flux=y_single*u.Jy, spectral_axis=x_single*u.um)

    # Fit the spectrum
    g_init = models.Gaussian1D(amplitude=3.*u.Jy, mean=6.1*u.um, stddev=1.*u.um)

    fit_params = {'maxiter': 200}

    g_fit = fit_lines(s_single, g_init, **fit_params)
    y_single_fit = g_fit(x_single*u.um)

    # Comparing every 10th value.
    y_single_fit_expected = np.array([3.69669474e-13, 3.57992454e-11, 2.36719426e-09, 1.06879318e-07,
               3.29498310e-06, 6.93605383e-05, 9.96945607e-04, 9.78431032e-03,
               6.55675141e-02, 3.00017760e-01, 9.37356842e-01, 1.99969007e+00,
               2.91286375e+00, 2.89719280e+00, 1.96758892e+00, 9.12412206e-01,
               2.88900005e-01, 6.24602556e-02, 9.22061121e-03, 9.29427266e-04]) * u.Jy

    assert np.allclose(y_single_fit.value[::10], y_single_fit_expected.value, atol=1e-5)


def test_spectrum_from_model():
    """
    This test fits the the first simulated spectrum from the fixture.  The
    initial guesses are manually set here with bounds that essentially make
    sense as the functionality of the test is to make sure the fit works and
    we get a reasonable answer out **given** good initial guesses.
    """

    np.random.seed(0)
    x = np.linspace(0., 10., 200)
    y = 3 * np.exp(-0.5 * (x - 6.3)**2 / 0.1**2)
    y += np.random.normal(0., 0.2, x.shape)

    y_continuum = 3.2 * np.exp(-0.5 * (x - 5.6)**2 / 4.8**2)
    y += y_continuum

    spectrum = Spectrum1D(flux=y*u.Jy, spectral_axis=x*u.um)

    # Unitless test
    chebyshev = models.Chebyshev1D(3, c0=0.1, c1=4, c2=5)
    spectrum_chebyshev = spectrum_from_model(chebyshev, spectrum)

    flux_expected = np.array([-4.90000000e+00, -3.64760991e-01, 9.22085553e+00, 2.38568496e+01,
                              4.35432211e+01, 6.82799702e+01, 9.80670968e+01, 1.32904601e+02,
                              1.72792483e+02, 2.17730742e+02, 2.67719378e+02, 3.22758392e+02,
                              3.82847784e+02, 4.47987553e+02, 5.18177700e+02, 5.93418224e+02,
                              6.73709126e+02, 7.59050405e+02, 8.49442062e+02, 9.44884096e+02])

    assert np.allclose(spectrum_chebyshev.flux.value[::10], flux_expected, atol=1e-5)

    # Unitfull test
    gaussian = models.Gaussian1D(amplitude=5*u.Jy, mean=4*u.um, stddev=2.3*u.um)
    spectrum_gaussian = spectrum_from_model(gaussian, spectrum)

    flux_expected = np.array([1.1020263, 1.57342489, 2.14175093, 2.77946243, 3.4389158,
                              4.05649712, 4.56194132, 4.89121902, 4.99980906, 4.872576,
                              4.52723165, 4.01028933, 3.3867847, 2.72689468, 2.09323522,
                              1.5319218, 1.06886794, 0.71101768, 0.45092638, 0.27264641])

    assert np.allclose(spectrum_gaussian.flux.value[::10], flux_expected, atol=1e-5)


def test_masking():
    """
    Test fitting spectra with masks
    """
    wl, flux = double_peak()
    s = Spectrum1D(flux=flux*u.Jy, spectral_axis=wl*u.um)

    # first we fit a single gaussian to the double_peak model, using the
    # known-good second peak (but a bit higher in amplitude). It should lock
    # in on the *second* peak since it's already close:
    g_init = models.Gaussian1D(2.5, 5.5, 0.2)
    g_fit1 = fit_lines(s, g_init)
    assert u.allclose(g_fit1.mean.value, 5.5, atol=.1)

    # now create a spectrum where the region around the second peak is masked.
    # The fit should now go to the *first* peak
    s_msk = Spectrum1D(flux=flux*u.Jy, spectral_axis=wl*u.um, mask=(5.1 < wl)&(wl < 6.1))
    g_fit2 = fit_lines(s_msk, g_init)
    assert u.allclose(g_fit2.mean.value, 4.6, atol=.1)

    # double check that it works with weights as well
    g_fit3 = fit_lines(s_msk, g_init, weights=np.ones_like(s_msk.flux.value))
    assert g_fit2.mean == g_fit3.mean


def test_window_extras():
    """
    Test that fitting works with masks and weights when a window is present
    """
    # similar to the masking test, but add a broad window around the whole thing
    wl, flux = double_peak()
    g_init = models.Gaussian1D(2.5, 5.5, 0.2)
    window_region = SpectralRegion(4*u.um, 8*u.um)
    mask = (5.1 < wl) & (wl < 6.1)

    s_msk = Spectrum1D(flux=flux*u.Jy, spectral_axis=wl*u.um, mask=mask)

    g_fit1 = fit_lines(s_msk, g_init, window=window_region)
    assert u.allclose(g_fit1.mean.value, 4.6, atol=.1)

    # check that if we weight instead of masking, we get the same result
    s = Spectrum1D(flux=flux*u.Jy, spectral_axis=wl*u.um)
    weights = (~mask).astype(float)
    g_fit2 = fit_lines(s, g_init, weights=weights, window=window_region)
    assert u.allclose(g_fit2.mean.value, 4.6, atol=.1)

    # and the same with both together
    weights = (~mask).astype(float)
    g_fit3 = fit_lines(s_msk, g_init, weights=weights, window=window_region)
    assert u.allclose(g_fit3.mean.value, 4.6, atol=.1)


def test_fit_subspectrum():
    # Create test spectrum
    spectrum = Spectrum1D(
        flux=np.random.sample(10) * u.Jy,
        spectral_axis=(np.arange(10) + 6558) * u.AA,
        rest_value=6561.16*u.angstrom,
        velocity_convention='relativistic')

    # Exract a sub region
    sub_region = SpectralRegion(6490 * u.AA, 6635 * u.AA)
    sub_spectrum = extract_region(spectrum, sub_region)

    # Create model and git to spectrum
    g_init = models.Gaussian1D(amplitude=1*u.Jy, mean=6561.157*u.angstrom,
                               stddev=4.0721*u.angstrom)
    with pytest.warns(AstropyUserWarning, match='The fit may be unsuccessful'):
        g_fit = fit_lines(sub_spectrum, g_init)
    yfit = g_fit(sub_spectrum.spectral_axis)  # noqa



import numpy as np

from scipy.interpolate import CubicSpline
from astropy.units import Quantity
from astropy.modeling import Fittable1DModel

from ..spectra import Spectrum1D
from ..utils import QuantityModel
from . import extract_region


def model_replace(spectrum, replace_region, model=10, extrapolation_treatment='data_fill',
                  interpolate_uncertainty=True):
    """
    Generates a new spectrum with a region replaced by a smooth spline.

    Parameters
    ----------

    spectrum : `~specutils.Spectrum1D`
        The spectrum to be modified.

    replace_region : `~specutils.SpectralRegion`
        Spectral region that specifies the region to be replaced. If None,
        parameter `model` below should define spline knots explicitly in the
        form of an `~astropy.units.Quantity` list or array

    model :
        An `~astropy.modeling` model object, which is assumed to have
        already been fit to the spectrum;
        Or
        a list or array of spectral axis values to be used as spline knots.
        They all should share the same units, which can be different
        from the units of the input spectrum spectral axis, but most
        be of compatible physical type;
        Or
        an integer value that will be used to build a list of equally-spaced
        knots, based on the `replace_region` instance.

    extrapolation_treatment : str
        What to do with data off the edges of the region encompassed by
        the spline knots. Default is ``'data_fill'`` to have points filled
        with the input flux values. ``'zero_fill'`` sets them to zero.

    interpolate_uncertainty : bool
        If True, the uncertainty, if present in the input spectrum, is
        also interpolated over the replaced segment.

    Returns
    -------
    spectrum : `~specutils.Spectrum1D`
        The spectrum with the region replaced by spline values, and data
        values or zeros outside the spline region. The spectral axis will
        have the same units as the spline knots.

    Examples
    --------

    >>> import numpy as np
    >>> import astropy.units as u
    >>> from specutils.spectra.spectrum1d import Spectrum1D
    >>> from specutils.manipulation.model_replace import model_replace
    >>> wave_val = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    >>> flux_val = np.array([2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
    >>> input_spectrum = Spectrum1D(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy)
    >>> spline_knots = [3.5, 4.7, 6.8, 7.1] * u.AA
    >>> result = model_replace(input_spectrum, None, spline_knots) # doctest: +IGNORE_OUTPUT

    """
    if extrapolation_treatment not in ('data_fill', 'zero_fill'):
        raise ValueError('invalid extrapolation_treatment value: ' +
                         str(extrapolation_treatment))

    # If input model is an array with spline knots and region
    # is not present, use knots directly to fit spline.
    if isinstance(model, Quantity) and replace_region is None:
        new_spectral_axis = spectrum.spectral_axis.to(model.unit)

        out_flux_val, out_uncert_val = _compute_spline_values(spectrum,
                                                              model,
                                                              new_spectral_axis,
                                                              interpolate_uncertainty)

    # If input model is an int, use it and the spectral region to build a
    # list of equally spaced spline knots.
    elif isinstance(model, int) and replace_region is not None:
        nknots = max(model, 3)
        dw = (replace_region.upper - replace_region.lower) / (nknots - 1)

        spline_knots = []
        for k in range(nknots):
            spline_knots.append(replace_region.lower + k * dw)

        spline_knots = spline_knots * replace_region.lower.unit
        new_spectral_axis = spectrum.spectral_axis.to(spline_knots.unit)

        out_flux_val, out_uncert_val = _compute_spline_values(spectrum,
                                                              spline_knots,
                                                              new_spectral_axis,
                                                              interpolate_uncertainty)

    # If input model is a fitted model, use it and the spectral region to place the
    # model values over the relevant stretch of the spectrum's spectral axis.
    elif (isinstance(model, Fittable1DModel) or (isinstance(model, QuantityModel))) \
            and replace_region is not None:
        new_spectral_axis = spectrum.spectral_axis

        subspectrum = extract_region(spectrum, replace_region)

        model_values = model(subspectrum.spectral_axis)

        out_flux_val = np.full(spectrum.spectral_axis.shape, np.nan)
        indices_up = np.where(spectrum.spectral_axis >= replace_region.lower)
        indices_dn = np.where(spectrum.spectral_axis <= replace_region.upper)
        i = int(indices_up[0][0])
        j = int(indices_dn[0][-1]) + 1

        out_flux_val[i:j] = model_values

        # models do not propagate uncertainties.
        out_uncert_val = None

    else:
        raise NotImplementedError("This combination of input parameters is not yet implemented.")

    # Careful with units handling from here on: astropylts handles the
    # np.where filter differently than the other supported environments.

    # Initialize with zero fill.
    out = np.where(np.isnan(out_flux_val), 0., out_flux_val) * spectrum.flux.unit

    # Fill extrapolated regions with original flux
    if extrapolation_treatment == 'data_fill':
        data = np.where(np.isnan(out_flux_val), spectrum.flux.value, 0.)
        out += (data * spectrum.flux.unit)

    # Do the same in case we want to interpolate the uncertainty.
    # Otherwise, do not propagate uncertainty into output.
    new_unc = None
    if out_uncert_val is not None:
        out_uncert = np.where(np.isnan(out_uncert_val), 0., out_uncert_val) * \
                     spectrum.uncertainty.unit
        data = np.where(np.isnan(out_uncert_val), spectrum.uncertainty.quantity.value, 0.)
        out_uncert += (data * spectrum.uncertainty.unit)

        new_unc = spectrum.uncertainty.__class__(array=out_uncert,
                                                 unit=spectrum.uncertainty.unit)

    return Spectrum1D(spectral_axis=new_spectral_axis, flux=out, uncertainty=new_unc)


def _compute_spline_values(spectrum, spline_knots, new_spectral_axis, interpolate_uncertainty):

    # Compute output flux values interpolated over the spline knots.
    out_flux_val = _interpolate_spline(spectrum.flux.value, new_spectral_axis, spline_knots)

    # Do the same in case we want to interpolate the uncertainty.
    # Otherwise, do not propagate uncertainty into output.
    out_uncert_val = None
    if spectrum.uncertainty is not None and interpolate_uncertainty:
        out_uncert_val = _interpolate_spline(spectrum.uncertainty.quantity,
                                             new_spectral_axis, spline_knots)

    return out_flux_val, out_uncert_val


def _interpolate_spline(input_values, spectral_axis, spline_knots):

    # Create spline to interpolate on input data.
    # Knots are the spectral axis values themselves.
    spline_1 = CubicSpline(spectral_axis.value, input_values,
                           extrapolate=False)

    # Now use that spline interpolator to compute interpolated
    # values from the input array at each spline knot.
    values_at_spline_knots = spline_1(spline_knots.value)

    # Finally, compute another spline interpolator over only the values
    # at spline knots, and use it to interpolate the output value at each
    # point on the spectral axis.
    spline_2 = CubicSpline(spline_knots.value, values_at_spline_knots,
                           extrapolate=False)

    return spline_2(spectral_axis.value)



"""
A module for analysis tools focused on determining the moment of
spectral features.
"""

import numpy as np
from ..manipulation import extract_region
from .utils import computation_wrapper


__all__ = ['moment']


def moment(spectrum, regions=None, order=0, axis=-1):
    """
    Estimate the moment of the spectrum.


    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum1d.Spectrum1D`
        The spectrum object over which the width will be calculated.

    regions: `~specutils.utils.SpectralRegion` or list of `~specutils.utils.SpectralRegion`
        Region within the spectrum to calculate the gaussian sigma width. If
        regions is `None`, computation is performed over entire spectrum.

    order : int
        The order of the moment to be calculated. Default=0

    axis : int
        Axis along which a moment is calculated. Default=-1, computes along
        the last axis (spectral axis).


    Returns
    -------
    moment: `float` or list (based on region input)
        Moment of the spectrum. Returns None if (order < 0 or None)

    """
    return computation_wrapper(_compute_moment, spectrum, regions,
                               order=order, axis=axis)


def _compute_moment(spectrum, regions=None, order=0, axis=-1):
    """
    This is a helper function for the above `moment()` method.
    """
    if regions is not None:
        calc_spectrum = extract_region(spectrum, regions)
    else:
        calc_spectrum = spectrum

    # Ignore masks for now. This should be fully addressed when
    # specutils gets revamped to handle multi-dimensional masks.
    flux = calc_spectrum.flux
    spectral_axis = calc_spectrum.spectral_axis

    if order is None or order < 0:
        return None

    if order == 0:
        return np.sum(flux, axis=axis)

    dispersion = spectral_axis
    if len(flux.shape) > len(spectral_axis.shape):
        _shape = flux.shape[:-1] + (1,)
        dispersion = np.tile(spectral_axis, _shape)

    if order == 1:
        return np.sum(flux * dispersion, axis=axis) / np.sum(flux, axis=axis)

    if order > 1:
        m0 = np.sum(flux, axis=axis)

        # By setting keepdims to True, the axes which are reduced are
        # left in the result as dimensions with size one. This means
        # that we can broadcast m1 correctly against dispersion.
        m1 = (np.sum(flux * spectral_axis, axis=axis, keepdims=True)
              / np.sum(flux, axis=axis, keepdims=True))

        return np.sum(flux * (dispersion - m1) ** order, axis=axis) / m0



"""
A module for internal utilities for the analysis sub-package.  Not meant for
public API consumption.
"""

from ..spectra import SpectralRegion


__all__ = ['computation_wrapper']


def computation_wrapper(func, spectrum, region, **kwargs):
    """
    Applies a computation across either a whole spectrum or a bunch of regions.
    """

    # No region, therefore whole spectrum.
    if region is None:
        return func(spectrum, **kwargs)

    # Single region
    elif isinstance(region, SpectralRegion):
        return func(spectrum, regions=region, **kwargs)

    # List of regions
    elif isinstance(region, list):
        return [func(spectrum, regions=reg, **kwargs) for reg in region]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Specutils: an astropy package for spectroscopy.
"""

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
from astropy import config as _config
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:  # noqa
    # For egg_info test builds to pass, put package imports here.

    # Allow loading spectrum object from top level module
    from .spectra import *  # noqa

    # Load the IO functions
    from .io.default_loaders import *  # noqa
    from .io.registers import _load_user_io
    _load_user_io()

__citation__ = 'https://doi.org/10.5281/zenodo.1421356'


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for specutils.
    """

    do_continuum_function_check = _config.ConfigItem(
        True,
        'Whether to check the spectrum baseline value is close'
        'to zero. If it is not within ``threshold`` then a warning is raised.'
    )


conf = Conf()



import numpy as np
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, InverseVariance

from ..manipulation import (FluxConservingResampler,
                            LinearInterpolatedResampler,
                            SplineInterpolatedResampler)
from ..spectra.spectrum1d import Spectrum1D

__all__ = ['template_match', 'template_redshift']


def _uncertainty_to_standard_deviation(uncertainty):
    """
    Convenience function to convert other uncertainty types to standard deviation,
    for consistency in calculations elsewhere.

    Parameters
    ----------
    uncertainty : :class:`~astropy.nddata.NDUncertainty`
        The input uncertainty

    Returns
    -------
    :class:`~numpy.ndarray`
        The array of standard deviation values.

    """
    if uncertainty is not None:
        if isinstance(uncertainty, StdDevUncertainty):
            stddev = uncertainty.array
        elif isinstance(uncertainty, VarianceUncertainty):
            stddev = np.sqrt(uncertainty.array)
        elif isinstance(uncertainty, InverseVariance):
            stddev = 1 / np.sqrt(uncertainty.array)

        return stddev


def _normalize_for_template_matching(observed_spectrum, template_spectrum, stddev=None):
    """
    Calculate a scale factor to be applied to the template spectrum so the
    total flux in both spectra will be the same.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum, which needs to be normalized in order to be
        compared with the observed spectrum.

    Returns
    -------
    `float`
        A float which will normalize the template spectrum's flux so that it
        can be compared to the observed spectrum.
    """
    if stddev is None:
        stddev = _uncertainty_to_standard_deviation(observed_spectrum.uncertainty)
    num = np.sum((observed_spectrum.flux*template_spectrum.flux) / (stddev**2))
    denom = np.sum((template_spectrum.flux / stddev)**2)

    return num/denom


def _resample(resample_method):
    """
    Find the user preferred method of resampling the template spectrum to fit
    the observed spectrum.

    Parameters
    ----------
    resample_method: `string`
        The type of resampling to be done on the template spectrum.

    Returns
    -------
    :class:`~specutils.ResamplerBase`
        This is the actual class that will handle the resampling.
    """
    if resample_method == "flux_conserving":
        return FluxConservingResampler()

    if resample_method == "linear_interpolated":
        return LinearInterpolatedResampler()

    if resample_method == "spline_interpolated":
        return SplineInterpolatedResampler()

    return None


def _chi_square_for_templates(observed_spectrum, template_spectrum, resample_method):
    """
    Resample the template spectrum to match the wavelength of the observed
    spectrum. Then, calculate chi2 on the flux of the two spectra.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum, which will be resampled to match the wavelength
        of the observed spectrum.

    Returns
    -------
    normalized_template_spectrum : :class:`~specutils.Spectrum1D`
        The normalized spectrum template.
    chi2 : `float`
        The chi2 of the flux of the observed spectrum and the flux of the
        normalized template spectrum.
    """
    # Resample template
    if _resample(resample_method) != 0:
        fluxc_resample = _resample(resample_method)
        template_obswavelength = fluxc_resample(template_spectrum,
                                                observed_spectrum.spectral_axis)

    # Convert the uncertainty to standard deviation if needed
    stddev = _uncertainty_to_standard_deviation(observed_spectrum.uncertainty)

    # Normalize spectra
    normalization = _normalize_for_template_matching(observed_spectrum,
                                                     template_obswavelength,
                                                     stddev)

    # Numerator
    num_right = normalization * template_obswavelength.flux
    num = observed_spectrum.flux - num_right

    # Denominator
    denom = stddev * observed_spectrum.flux.unit

    # Get chi square
    result = (num/denom)**2
    chi2 = np.sum(result.value)

    # Create normalized template spectrum, which will be returned with
    # corresponding chi2
    normalized_template_spectrum = Spectrum1D(
        spectral_axis=template_spectrum.spectral_axis,
        flux=template_spectrum.flux*normalization)

    return normalized_template_spectrum, chi2


def template_match(observed_spectrum, spectral_templates,
                   resample_method="flux_conserving",
                   redshift=None):
    """
    Find which spectral templates is the best fit to an observed spectrum by
    computing the chi-squared. If two template_spectra have the same chi2, the
    first template is returned.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    spectral_templates : :class:`~specutils.Spectrum1D` or :class:`~specutils.SpectrumCollection` or `list`
        That will give a single :class:`~specutils.Spectrum1D` when iterated
        over. The template spectra, which will be resampled, normalized, and
        compared to the observed spectrum, where the smallest chi2 and
        normalized template spectrum will be returned.
    resample_method : `string`
        Three resample options: flux_conserving, linear_interpolated, and spline_interpolated.
        Anything else does not resample the spectrum.
    redshift : 'float', `int`, `list`, `tuple`, 'numpy.array`
        If the user knows the redshift they want to apply to the spectrum/spectra within spectral_templates,
        then this float or int value redshift can be applied to each template before attempting the match.
        Or, alternatively, an iterable with redshift values to be applied to each template, before computation
        of the corresponding chi2 value, can be passed via this same parameter. For each template, the redshift
        value that results in the smallest chi2 is used.

    Returns
    -------
    normalized_template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum that has been normalized.
    redshift : `None` or `float`
        The value of the redshift that provides the smallest chi2.
    smallest_chi_index : `int`
        The index of the spectrum with the smallest chi2 in spectral templates.
    chi2_min : `float`
        The chi2 of the flux of the observed_spectrum and the flux of the
        normalized template spectrum.
    chi2_list : `list`
        A list with all chi2 values found for each template spectrum.
    """
    final_redshift = None

    if hasattr(spectral_templates, 'flux') and len(spectral_templates.flux.shape) == 1:

        # Account for redshift if provided
        chi2_list = []
        if redshift is not None:
            results = template_redshift(observed_spectrum, spectral_templates, redshift=redshift)
            redshifted_spectrum, final_redshift, normalized_spectral_template, chi2, chi2_inner_list = results # noqa

        else:
            normalized_spectral_template, chi2 = _chi_square_for_templates(
                observed_spectrum, spectral_templates, resample_method)

        chi2_list.append(chi2)

        return normalized_spectral_template, final_redshift, 0, chi2, chi2_list

    # At this point, the template spectrum is either a ``SpectrumCollection``
    # or a multi-dimensional``Spectrum1D``. Loop through the object and return
    # the template spectrum with the lowest chi square and its corresponding
    # chi square.
    chi2_min = None
    smallest_chi_spec = None
    chi2_list = []

    final_redshift = None
    temp_redshift = None

    for index, spectrum in enumerate(spectral_templates):

        # Account for redshift if provided
        if redshift is not None:
            results = template_redshift(observed_spectrum, spectrum, redshift=redshift)
            redshifted_spectrum, temp_redshift, normalized_spectral_template, chi2, chi2_inner_list = results # noqa

            chi2_list.append(chi2_inner_list)

        else:
            normalized_spectral_template, chi2 = _chi_square_for_templates(
                observed_spectrum, spectrum, resample_method)

        if chi2_min is None or chi2 < chi2_min:
            chi2_min = chi2
            smallest_chi_spec = normalized_spectral_template
            smallest_chi_index = index

            final_redshift = temp_redshift

    return smallest_chi_spec, final_redshift, smallest_chi_index, chi2_min, chi2_list


def template_redshift(observed_spectrum, template_spectrum, redshift):
    """
    Find the best-fit redshift for template_spectrum to match observed_spectrum using chi2.

    Parameters
    ----------
    observed_spectrum : :class:`~specutils.Spectrum1D`
        The observed spectrum.
    template_spectrum : :class:`~specutils.Spectrum1D`
        The template spectrum, which will have it's redshift calculated.
    redshift : `float`, `int`, `list`, `tuple`, 'numpy.array`
        A scalar or iterable with the redshift values to test.

    Returns
    -------
    redshifted_spectrum: :class:`~specutils.Spectrum1D`
        A new Spectrum1D object which incorporates the template_spectrum with a spectral_axis
        that has been redshifted using the final_redshift.
    final_redshift : `float`
        The best-fit redshift for template_spectrum to match the observed_spectrum.
    normalized_template_spectrum : :class:`~specutils.Spectrum1D`
        The normalized spectrum template.
    chi2_min: `float`
        The smallest chi2 value that was found.
    chi2_list : `list`
        A list with the chi2 values corresponding to each input redshift value.
    """
    chi2_min = None
    final_redshift = None
    chi2_list = []

    redshift = np.array(redshift).reshape((np.array(redshift).size,))

    # Loop which goes through available redshift values and finds the smallest chi2
    for rs in redshift:

        # Create new redshifted spectrum and run it through the chi2 method
        redshifted_spectrum = Spectrum1D(spectral_axis=template_spectrum.spectral_axis*(1+rs),
                                         flux=template_spectrum.flux,
                                         uncertainty=template_spectrum.uncertainty,
                                         meta=template_spectrum.meta)

        normalized_spectral_template, chi2 = _chi_square_for_templates(
            observed_spectrum, redshifted_spectrum, "flux_conserving")

        chi2_list.append(chi2)

        # Set new chi2_min if suitable replacement is found
        if not np.isnan(chi2) and (chi2_min is None or chi2 < chi2_min):
            chi2_min = chi2
            final_redshift = rs

    return redshifted_spectrum, final_redshift, normalized_spectral_template, chi2_min, chi2_list
